"""
Note that backend.release_vessel() calls that automatically happen when not 
enough of the requested vessels are acquired don't currently need to be
mock'd out because currently that function does nothing.
"""

# The seattlegeni testlib must be imported first.
from seattlegeni.tests import testlib

from seattlegeni.tests import mocklib

from seattlegeni.common.api import maindb

from seattlegeni.common.exceptions import *

from seattlegeni.website.control import interface

from seattlegeni.website.tests import testutil

import unittest





mocklib.mock_lockserver_calls()





class SeattleGeniTestCase(unittest.TestCase):


  def setUp(self):
    # Setup a fresh database for each test.
    testlib.setup_test_db()



  def tearDown(self):
    # Cleanup the test database.
    testlib.teardown_test_db()



  def test_acquire_vessels_invalid_request(self):
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    
    func = interface.acquire_vessels
    
    # Negative vesselcount.
    args = (user, -1, 'wan')
    self.assertRaises(AssertionError, func, *args)   

    # Zero vesselcount.
    args = (user, 0, 'wan')
    self.assertRaises(AssertionError, func, *args)   

    # Unrecognized vessel type.
    args = (user, 1, 'x')
    self.assertRaises(ProgrammerError, func, *args)   



  def test_acquire_vessels_insufficient_vessel_credits(self):
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    
    # TODO: need to test maindb.require_user_can_acquire_resources separately
    
    # The user doesn't have any donations, they shouldn't be able to acquire
    # more vessels than their free credits.
    credit_limit = maindb.get_user_free_vessel_credits(user)
    
    vesseltypelist = ['wan', 'lan', 'nat', 'rand']

    for vesseltype in vesseltypelist:
      func = interface.acquire_vessels
      args = (user, credit_limit + 1, vesseltype)
      self.assertRaises(InsufficientUserResourcesError, func, *args)      
  


  def test_acquire_wan_vessels_multiple_calls_wan(self):

    # Have every vessel acquisition to the backend request succeed.
    calls_results = [True] * 10
    mocklib.mock_backend_acquire_vessel(calls_results)
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    userport = user.usable_vessel_port
    
    vesselcount = maindb.get_user_free_vessel_credits(user)
    
    testutil.create_nodes_on_different_subnets(vesselcount, [userport])
    
    # First request a single vessel.
    first_vessel_list = interface.acquire_vessels(user, 1, 'wan')
    
    # Now acquire all of the rest that the user can acquire.
    second_vessel_list = interface.acquire_vessels(user, vesselcount - 1, 'wan')

    self.assertEqual(1, len(first_vessel_list))
    self.assertEqual(vesselcount - 1, len(second_vessel_list))



  def test_acquire_wan_vessels_multiple_calls_lan(self):

    # Have every vessel acquisition to the backend request succeed.
    calls_results = [True] * 10
    mocklib.mock_backend_acquire_vessel(calls_results)
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    userport = user.usable_vessel_port
    
    vesselcount = maindb.get_user_free_vessel_credits(user)
    
    testutil.create_nodes_on_same_subnet(vesselcount, [userport])
    
    # First request a single vessel.
    first_vessel_list = interface.acquire_vessels(user, 1, 'lan')
    
    # Now acquire all of the rest that the user can acquire.
    second_vessel_list = interface.acquire_vessels(user, vesselcount - 1, 'lan')

    self.assertEqual(1, len(first_vessel_list))
    self.assertEqual(vesselcount - 1, len(second_vessel_list))



  def test_acquire_wan_vessels_multiple_calls_nat(self):

    # Have every vessel acquisition to the backend request succeed.
    calls_results = [True] * 10
    mocklib.mock_backend_acquire_vessel(calls_results)
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    userport = user.usable_vessel_port
    
    vesselcount = maindb.get_user_free_vessel_credits(user)
    
    testutil.create_nat_nodes(vesselcount, [userport])
    
    # First request a single vessel.
    first_vessel_list = interface.acquire_vessels(user, 1, 'nat')
    
    # Now acquire all of the rest that the user can acquire.
    second_vessel_list = interface.acquire_vessels(user, vesselcount - 1, 'nat')

    self.assertEqual(1, len(first_vessel_list))
    self.assertEqual(vesselcount - 1, len(second_vessel_list))



  def test_acquire_wan_vessels_multiple_calls_rand(self):

    # Have every vessel acquisition to the backend request succeed.
    calls_results = [True] * 10
    mocklib.mock_backend_acquire_vessel(calls_results)
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    userport = user.usable_vessel_port
    
    vesselcount = maindb.get_user_free_vessel_credits(user)
    
    # Create vesselcount nodes split between the different types.
    testutil.create_nodes_on_different_subnets(4, [userport])
    testutil.create_nodes_on_same_subnet(4, [userport])
    testutil.create_nat_nodes(vesselcount - 8, [userport])
    
    # First request a single vessel.
    first_vessel_list = interface.acquire_vessels(user, 1, 'rand')
    
    # Now acquire all of the rest that the user can acquire.
    second_vessel_list = interface.acquire_vessels(user, vesselcount - 1, 'rand')

    self.assertEqual(1, len(first_vessel_list))
    self.assertEqual(vesselcount - 1, len(second_vessel_list))



  def test_acquire_wan_vessels_some_vessels_fail(self):

    # Have every other vessel acquisition fail. We're going to acquire 50,
    # so we'll need 100 responses alternating between failure and success
    # (we're starting with failure, so 100, not 99).
    calls_results = [False, True] * 50
    mocklib.mock_backend_acquire_vessel(calls_results)
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    userport = user.usable_vessel_port
    
    # We need to give the user some donations so they have enough credits.
    # We're assuming it's 10 credits per donation.
    self.assertEqual(10, maindb.VESSEL_CREDITS_FOR_DONATIONS_MULTIPLIER)
    
    # Also make sure the user started with 10 credits.
    self.assertEqual(10, maindb.get_user_free_vessel_credits(user))
    
    # We need 100 nodes the user can acquire vessels on as we're having half of
    # the node acquisitions fail.
    testutil.create_nodes_on_different_subnets(100, [userport])
    
    # Now credit the user for donations on 4 of these.
    for node in maindb.get_active_nodes()[:4]:
      maindb.create_donation(node, user, '')
    
    # Ok, the user now has 50 vessels the can acquire and there are 100 nodes
    # with vessels available for them. Let's try to acquire all 50 at once and
    # make sure this works even though we'll have to get through 100 requests
    # to the backend to make it happen.
    vessel_list = interface.acquire_vessels(user, 50, 'wan')
    
    self.assertEqual(50, len(vessel_list))




  def test_acquire_lan_vessels_some_subnets_have_too_many_vessels_fail(self):
    """
    We're going to be trying to acquire 4 vessels in a single subnet. We'll
    make it so that there are 3 subnets to choose from. The first two will
    potentially have enough vessels for the user to satisfy their acquisition
    request, but too many vessels in each will fail. The third subnet tried
    will have enough succeed.
    """
    
    # We're going to have three subnets with 5 potential vessels each. We
    # want the first two subnets to fail. Choosing the values creatively
    # here is helped by knowing the logic of the function
    # vessels._acquire_vessels_from_list(). Basically, it won't try to
    # acquire more than the minumum needed at a time, so the first time
    # it loops it will try to acquire all 4, then the next time it loops
    # it will try to acquire however many of the 4 failed the first time.
    # It won't bother trying again if there are too few left in the list.
    results_1 = [False, True, True, True, False]
    results_2 = [False, False, True, True]
    results_3 = [False, True, True, True, True]
    calls_results = results_1 + results_2 + results_3
    mocklib.mock_backend_acquire_vessel(calls_results)
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    userport = user.usable_vessel_port
    
    # Make sure the user starts with 10 credits.
    self.assertEqual(10, maindb.get_user_free_vessel_credits(user))
    
    # Create three subnets with 5 nodes each. We need to keep them all the same
    # size, otherwise we need to change the test to patch maindb so that the
    # subnets will be tried in the order we expect.
    testutil.create_nodes_on_same_subnet(5, [userport], ip_prefix="127.0.0.")
    testutil.create_nodes_on_same_subnet(5, [userport], ip_prefix="127.0.1.")
    testutil.create_nodes_on_same_subnet(5, [userport], ip_prefix="127.0.2.")
    
    # Now try to acquire 8 lan nodes on a single subnet.
    vessel_list = interface.acquire_vessels(user, 4, 'lan')
    
    self.assertEqual(4, len(vessel_list))
    
    # Make sure backend.acquire_vessel() got called the correct number of
    # times (that is, the call_results list got pop(0)'d enough times).
    self.assertEqual(0, len(calls_results))



  def test_acquire_specific_vessels(self):
    # 8 vessels will ultimately be acquired.
    calls_results = [True] * 8
    mocklib.mock_backend_acquire_vessel(calls_results)
    
    # Create a user who will be doing the acquiring.
    user = maindb.create_user("testuser", "password", "example@example.com", "affiliation", "1 2", "2 2 2", "3 4")
    userport = user.usable_vessel_port
    
    vesselcount = maindb.get_user_free_vessel_credits(user)
    
    # We use userport + 1 to make sure the user isn't being restricted to only
    # vessels that have their user port in their port list.
    testutil.create_nodes_on_different_subnets(vesselcount + 10, [userport + 1])
    
    vessels = list(maindb._get_queryset_of_all_available_vessels_for_a_port_include_nat_nodes(userport + 1))
    
    # Request the first 4 vessels in the list.
    first_vessel_list = interface.acquire_specific_vessels(user, vessels[:4])
    
    # Now request the first 6 vessels in the list. We should only get 2.
    second_vessel_list = interface.acquire_specific_vessels(user, vessels[:6])

    self.assertEqual(4, len(first_vessel_list))
    self.assertEqual(2, len(second_vessel_list))
    
    # Now ask for more vessels than the user has available, regardless of the
    # fact that some of the requested vessels aren't available.
    requestcount = vesselcount - 6 + 1
    
    func = interface.acquire_specific_vessels
    args = (user, vessels[:requestcount])
    self.assertRaises(InsufficientUserResourcesError, func, *args)





def run_test():
  unittest.main()



if __name__ == "__main__":
  run_test()



import unittest

import lockserver_daemon as lockserver


class TheTestCase(unittest.TestCase):

  def setUp(self):
    # Reset the lockserver's global variables between each test.
    lockserver.init_globals()


  def testContentionForSingleUserLock(self):
    # This entire test will use this same lockdict in each request
    locks = {'user':['bob']}
    
    # Start three sessions.
    sess = []
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())

    # Have all three session try to acquire a lock on a single user.
    lockserver.do_acquire_locks(sess[0], locks)
    lockserver.do_acquire_locks(sess[1], locks)
    lockserver.do_acquire_locks(sess[2], locks)
    
    # The first session should have the lock and the other two sessions should
    # be queued in order for that lock.
    expected_heldlockdict = {
      'node': {},
      'user': {'bob': {'locked_by_session': sess[0],
                       'queue': [sess[1], sess[2]]}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'node': [], 'user': ['bob']},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True},
      sess[1]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': ['bob']},
                'acquirelocksproceedeventset': False},
      sess[2]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': ['bob']},
                'acquirelocksproceedeventset': False}}
    
    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])
    
    # Have the session that holds the lock release it.
    lockserver.do_release_locks(sess[0], locks)
    
    # The first session should have the lock and the other two sessions should
    # be queued in order for that lock.
    expected_heldlockdict = {
      'node': {},
      'user': {'bob': {'locked_by_session': sess[1],
                       'queue': [sess[2]]}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True},
      sess[1]: {'heldlocks': {'node': [], 'user': ['bob']},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True},
      sess[2]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': ['bob']},
                'acquirelocksproceedeventset': False}}
    
    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])
  
    # The session that just released it tries to acquire it again.
    lockserver.do_acquire_locks(sess[0], locks)
    
    # The first session should have the lock and the other two sessions should
    # be queued in order for that lock.
    expected_heldlockdict = {
      'node': {},
      'user': {'bob': {'locked_by_session': sess[1],
                       'queue': [sess[2], sess[0]]}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': ['bob']},
                'acquirelocksproceedeventset': False},
      sess[1]: {'heldlocks': {'node': [], 'user': ['bob']},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True},
      sess[2]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': ['bob']},
                'acquirelocksproceedeventset': False}}
    
    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])


  def testContentionForSingleNodeLock(self):
    """
    This is the same as testContentionForSingleUserLock but involves contention
    over a single node lock rather than over a single user lock. Both could have
    been done with much less code using a helper method, but that makes the
    dictionaries of expected data much less intuitive to look at and thus
    makes the tests hard to understand. So, just duplicating code, instead. 
    """
    # This entire test will use this same lockdict in each request
    locks = {'node':['123']}
    
    # Start three sessions.
    sess = []
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())

    # Have all three session try to acquire a lock on a single user.
    lockserver.do_acquire_locks(sess[0], locks)
    lockserver.do_acquire_locks(sess[1], locks)
    lockserver.do_acquire_locks(sess[2], locks)
    
    # The first session should have the lock and the other two sessions should
    # be queued in order for that lock.
    expected_heldlockdict = {
      'user': {},
      'node': {'123': {'locked_by_session': sess[0],
                       'queue': [sess[1], sess[2]]}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'user': [], 'node': ['123']},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True},
      sess[1]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': [], 'node': ['123']},
                'acquirelocksproceedeventset': False},
      sess[2]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': [], 'node': ['123']},
                'acquirelocksproceedeventset': False}}
    
    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])
    
    # Have the session that holds the lock release it.
    lockserver.do_release_locks(sess[0], locks)
    
    # The first session should have the lock and the other two sessions should
    # be queued in order for that lock.
    expected_heldlockdict = {
      'user': {},
      'node': {'123': {'locked_by_session': sess[1],
                       'queue': [sess[2]]}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True},
      sess[1]: {'heldlocks': {'user': [], 'node': ['123']},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True},
      sess[2]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': [], 'node': ['123']},
                'acquirelocksproceedeventset': False}}
    
    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])
  
    # The session that just released it tries to acquire it again.
    lockserver.do_acquire_locks(sess[0], locks)
    
    # The first session should have the lock and the other two sessions should
    # be queued in order for that lock.
    expected_heldlockdict = {
      'user': {},
      'node': {'123': {'locked_by_session': sess[1],
                       'queue': [sess[2], sess[0]]}}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': [], 'node': ['123']},
                'acquirelocksproceedeventset': False},
      sess[1]: {'heldlocks': {'user': [], 'node': ['123']},
                'neededlocks': {'user': [], 'node': []},
                'acquirelocksproceedeventset': True},
      sess[2]: {'heldlocks': {'user': [], 'node': []},
                'neededlocks': {'user': [], 'node': ['123']},
                'acquirelocksproceedeventset': False}}
    
    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])




#!/usr/bin/env python
"""
This script is used to output single lines of comma-separated data for graphing
statistics related to seattlegeni.

Usage:
  python print_data_point.py node_overview CURRENT_VERSION
  or
  python print_data_point.py node_type
  or
  python print_data_point.py vessels
  or
  python print_data_point.py advertise

  When using the 'node_overview' option, you have to pass an extra argument of
  the current version of seattlegeni. For example:
    python print_data_point.py node_overview 0.1n

All of the usages print out something like this, just one single line:

2009-10-29 14:30,472,3

The above is a date (2009-10-29 14:30), followed by two numbers (472 and 3).
"""

from seattlegeni.website.control.models import Node
from seattlegeni.website.control.models import Vessel
from seattlegeni.website import settings

from seattlegeni.common.api import maindb

import datetime
import sys

# Import necessary repy files
from seattle.repyportability import *
add_dy_support(locals())

dy_import_module_symbols("rsa.r2py")
dy_import_module_symbols("advertise.r2py")

# This is the v2.publickey, the key is advertised by every single node manager thats up and running.
v2key = {'e': 22599311712094481841033180665237806588790054310631222126405381271924089573908627143292516781530652411806621379822579071415593657088637116149593337977245852950266439908269276789889378874571884748852746045643368058107460021117918657542413076791486130091963112612854591789518690856746757312472362332259277422867L, 'n': 12178066700672820207562107598028055819349361776558374610887354870455226150556699526375464863913750313427968362621410763996856543211502978012978982095721782038963923296750730921093699612004441897097001474531375768746287550135361393961995082362503104883364653410631228896653666456463100850609343988203007196015297634940347643303507210312220744678194150286966282701307645064974676316167089003178325518359863344277814551559197474590483044733574329925947570794508677779986459413166439000241765225023677767754555282196241915500996842713511830954353475439209109249856644278745081047029879999022462230957427158692886317487753201883260626152112524674984510719269715422340038620826684431748131325669940064404757120601727362881317222699393408097596981355810257955915922792648825991943804005848347665699744316223963851263851853483335699321871483966176480839293125413057603561724598227617736944260269994111610286827287926594015501020767105358832476708899657514473423153377514660641699383445065369199724043380072146246537039577390659243640710339329506620575034175016766639538091937167987100329247642670588246573895990251211721839517713790413170646177246216366029853604031421932123167115444834908424556992662935981166395451031277981021820123445253L}




def main():
  arg_to_func_dict = {"node_overview" : get_node_overview_line, 
                      "node_type": get_node_type_line, 
                      "vessels" : get_vessels_line,
                      "advertise" : get_advertise_line}
    
  if len(sys.argv) < 2 or sys.argv[1] not in arg_to_func_dict:
    print "Usage: print_data_points.py [" + "|".join(arg_to_func_dict.keys()) + "] (possible other args)" 
    sys.exit(1)

  print arg_to_func_dict[sys.argv[1]]()





def _datestr():
  date = datetime.datetime.now()
  return "%d-%02d-%02d %02d:%02d" % (date.year, date.month, date.day,
                                     date.hour, date.minute)





def get_node_overview_line():
  parts = []
  parts.append(_datestr())
  parts.append(str(_active_node_count()))
  parts.append(str(_active_broken_node_count()))
  parts.append(str(_active_old_version_node_count()))
  return ",".join(parts)


def _active_node_count():
  return len(maindb.get_active_nodes())


def _active_broken_node_count():
  return len(maindb.get_active_nodes_include_broken()) - len(maindb.get_active_nodes())


def _active_old_version_node_count():
  currentversion = sys.argv[2]
  queryset = Node.objects.filter(is_active=True, is_broken=False)
  queryset = queryset.exclude(last_known_version=currentversion)
  return queryset.count()





def get_node_type_line():
  parts = []
  parts.append(_datestr())
  parts.append(str(_active_non_nat_node_count()))
  parts.append(str(_active_nat_node_count()))
  return ",".join(parts)


def _active_non_nat_node_count():
  queryset = Node.objects.filter(is_active=True, is_broken=False)
  queryset = queryset.exclude(last_known_ip__startswith=maindb.NAT_STRING_PREFIX)
  return queryset.count()


def _active_nat_node_count():
  queryset = Node.objects.filter(is_active=True, is_broken=False)
  queryset = queryset.filter(last_known_ip__startswith=maindb.NAT_STRING_PREFIX)
  return queryset.count()





def get_vessels_line():
  parts = []
  parts.append(_datestr())
  parts.append(str(_free_vessels()))
  parts.append(str(_acquired_vessels()))
  parts.append(str(_dirty_vessels()))
  return ",".join(parts)


def _free_vessels():
  queryset = Vessel.objects.filter(acquired_by_user=None)
  queryset = queryset.filter(node__is_active=True, node__is_broken=False)
  return queryset.count()


def _acquired_vessels():
  queryset = Vessel.objects.exclude(acquired_by_user=None)
  queryset = queryset.exclude(date_expires__lte=datetime.datetime.now())
  return queryset.count()


def _dirty_vessels():
  return len(maindb.get_vessels_needing_cleanup())



def _state_key_file_to_publickey(key_file_name):
  """ Retrieve pubkey from file and return the dictionary form of key"""
  return rsa_file_to_publickey(os.path.join(settings.SEATTLECLEARINGHOUSE_STATE_KEYS_DIR, key_file_name))



def get_advertise_line():
  """
  <Purpose>
    Lookup the nodes with their transition state key in order
    to find out how many nodes are in each state. Then print the
    number of nodes that are in each state.

  <Arguments>
    None

  <Exception>
    None

  <Side Effects>
    None
  
  <Return>
    None
  """

  state_keys = {"canonical" : _state_key_file_to_publickey("canonical.publickey"),
                "acceptdonation" : _state_key_file_to_publickey("acceptdonation.publickey"),
                "movingto_onepercentmanyevents" : _state_key_file_to_publickey("movingto_onepercentmanyevents.publickey"),
                "onepercent_manyevents" : _state_key_file_to_publickey("onepercentmanyevents.publickey")}

 
  parts = []
  parts.append(_datestr())
  parts.append(str(_lookup_nodes(state_keys["acceptdonation"])))
  parts.append(str(_lookup_nodes(state_keys["canonical"])))
  parts.append(str(_lookup_nodes(state_keys["movingto_onepercentmanyevents"])))
  parts.append(str(_lookup_nodes(state_keys["onepercent_manyevents"])))
  parts.append(str(_lookup_nodes(v2key)))
  return ",".join(parts)


def _lookup_nodes(node_state_pubkey):
  """
  <Purpose>
    Lookup nodes given a publickey.

  <Arguments>
    node_state_pubkey - publickey to use to lookup nodes.

  <Exception>
    Exception is printed to stderr.

  <Side Effects>
    None

  <Return>
    The number of nodes found using the public key.
  """

  # Lookup nodes using publickey.   If it timesout or fails, this is okay.  
  # just log it, return -1 (to print that in the log) and continue
  try:
    node_list = advertise_lookup(node_state_pubkey, maxvals = 10*1024*1024)
  except Exception, e:
    print >> sys.stderr, e
    return -1
  
  return len(node_list)


    
if __name__ == "__main__":
  main()





import unittest

import lockserver_daemon as lockserver


class TheTestCase(unittest.TestCase):

  def setUp(self):
    # Reset the lockserver's global variables between each test.
    lockserver.init_globals()


  def testStartSessionCreatesNonemptyStringsForIds(self):
    sess = []
    sess.append(lockserver.do_start_session())
    
    self.assertNotEqual(sess[0], '')
    self.assertTrue(isinstance(sess[0], str))


  def testStartSessionCreatesUniqueIds(self):
    sess = []
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    
    self.assertNotEqual(sess[0], sess[1])
    self.assertNotEqual(sess[1], sess[2])
    
    
  def testEndSession(self):
    sess = []
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    
    expected_heldlockdict = {
      'node': {},
      'user': {}}
    expected_sessiondict = {
      sess[0]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True},
      sess[1]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True},
      sess[2]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True}}
    
    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])
    
    # End the first session.
    lockserver.do_end_session(sess[0])
    
    expected_heldlockdict = {
      'node': {},
      'user': {}}
    expected_sessiondict = {
      sess[1]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True},
      sess[2]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True}}
    
    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])
    
    # Make sure we aren't allowed to use the session we just ended in a lock request.
    locks = {'user':['bob']}
    func = lockserver.do_acquire_locks
    args = (sess[0], locks)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)
    
    # Make sure we aren't allowed to end the same session again.
    func = lockserver.do_end_session
    args = (sess[0],)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)
    
    # Make sure the sessions we didn't end still work (the second session
    # will acquire the lock, the third session will be queued for it).
    locks = {'user':['bob']}
    lockserver.do_acquire_locks(sess[1], locks)
    lockserver.do_acquire_locks(sess[2], locks)
    
    expected_heldlockdict = {
      'node': {},
      'user': {'bob': {'locked_by_session': sess[1],
                       'queue': [sess[2]]}}}
    expected_sessiondict = {
      sess[1]: {'heldlocks': {'node': [], 'user': ['bob']},
                'neededlocks': {'node': [], 'user': []},
                'acquirelocksproceedeventset': True},
      sess[2]: {'heldlocks': {'node': [], 'user': []},
                'neededlocks': {'node': [], 'user': ['bob']},
                'acquirelocksproceedeventset': False}}
    
    status = lockserver.do_get_status()
    self.assertEqual(expected_heldlockdict, status["heldlockdict"])
    self.assertEqual(expected_sessiondict, status["sessiondict"])


  def testEndSessionFailsWhenSessionNotEmpty(self):
    sess = []
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    sess.append(lockserver.do_start_session())
    
    # Have the first session acquire a lock.
    locks = {'user':['bob']}
    lockserver.do_acquire_locks(sess[0], locks)
    
    # Make sure we can't end the session while the session holds locks.
    func = lockserver.do_end_session
    args = (sess[0],)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)
  
    # Have the second session be queued for the same lock.
    locks = {'user':['bob']}
    lockserver.do_acquire_locks(sess[1], locks)
    
    # Make sure we can't end the session while the session has pending
    # lock acquisition requests.
    func = lockserver.do_end_session
    args = (sess[0],)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)


  def testNonexistentSessionId(self):
    
    sess = []
    sess.append('1234')
    
    # Try to acquire locks with the non-existent session.
    locks = {'user':['bob']}
    func = lockserver.do_acquire_locks
    args = (sess[0], locks)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)
    
    # Try to release locks with the non-existent session.
    locks = {'user':['bob']}
    func = lockserver.do_acquire_locks
    args = (sess[0], locks)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)
  
    # Try to end the non-existent session.
    func = lockserver.do_end_session
    args = (sess[0],)
    self.assertRaises(lockserver.LockserverInvalidRequestError, func, *args)
  



"""
This file must be readable by the backend and should only be readable by client
code that is allowed to make privileged requests to the backend. Basically,
it should *not* be readable by the user the webserver/website runs as.

The reason for the restrictions is explained in backend_daemon.py, but
basically it's that we want the backend to be the place where all node-
"""

# Change this for production usage.
authcode = "FILL_THIS_IN"



from django.conf.urls.defaults import *

urlpatterns = patterns('seattlegeni.website.reports.views',

                       (r'^$', 'index', {}, 'index'),
                       (r'^all$', 'all', {}, 'all'),
                       (r'^donations$', 'donations', {}, 'donations'),
                       (r'^acquired_vessels$', 'acquired_vessels', {}, 'acquired_vessels'),
                       (r'^vessels_by_port$', 'vessels_by_port', {}, 'vessels_by_port'),
                       (r'^lan_sizes_by_port$', 'lan_sizes_by_port', {}, 'lan_sizes_by_port'),
                      )



#!/usr/bin/env python
"""
<Purpose>
  This script is used to immediately stop everything running on behalf of a given
  seattle user, remove the resources allocated to them, and restrict them from
  acquiring more resources.

  Usage:
    ./ban_user_and_stop_vessels.py <geni_username>

  You will be given status information as the process is completed, including
  any nodes that can't be communicated with (meaning that the user's vessels
  on that node haven't been taken away from them).
"""

import sys
import time

from seattlegeni.common.api import lockserver
from seattlegeni.common.api import maindb
from seattlegeni.common.util import log
from seattlegeni.website.control import vessels

from seattlegeni.common.exceptions import *


# Set the log level high enough so that we don't produce a bunch of logging
# output due to the logging decorators.
initial_log_level = log.loglevel
log.set_log_level(log.LOG_LEVEL_INFO)



def ban_user_and_remove_vessels(username):

  try:
    geniuser = maindb.get_user(username, allow_inactive=True)
  except DoesNotExistError:
    print "No such user: %s." % username
    sys.exit(1)

  # Lock the user.
  lockserver_handle = lockserver.create_lockserver_handle()
  lockserver.lock_user(lockserver_handle, geniuser.username)

  try:
    if geniuser.is_active:
      geniuser.is_active = False
      geniuser.save()
      print "This account has been set to inactive (banned)."
    else:
      print "This account is already inactive (banned)."

    acquired_vessels = maindb.get_acquired_vessels(geniuser)
    if not acquired_vessels:
      print "No acquired vessels to stop/remove access to."
    else:
      print "Vessels acquired by this user: %s" % acquired_vessels
      print "Indicating to the backend to release %s vessels." % len(acquired_vessels)
      vessels.release_vessels(lockserver_handle, geniuser, acquired_vessels)
      print "Release indicated. Monitoring db to see if the backend cleaned them up."
      while True:
        for vessel in acquired_vessels[:]:
          updated_vessel = maindb.get_vessel(vessel.node.node_identifier, vessel.name)
          if vessel.node.is_broken or not vessel.node.is_active:
            print "Node %s is broken or inactive, so backend won't contact it." % vessel.node
            acquired_vessels.remove(vessel)
            continue

          if updated_vessel.is_dirty:
            print "Vessel %s has not been cleaned up yet." % updated_vessel
          else:
            print "Vessel %s has been cleaned up." % updated_vessel
            acquired_vessels.remove(vessel)

        if not acquired_vessels:
          print "All vessels have been cleaned up."
          break
        else:
          print "%s vessels remain to be cleaned up." % len(acquired_vessels)

        print "Sleeping 10 seconds."
        time.sleep(10)

  finally:
    # Unlock the user.
    lockserver.unlock_user(lockserver_handle, geniuser.username)
    lockserver.destroy_lockserver_handle(lockserver_handle)


if __name__ == "__main__":
  if len(sys.argv) != 2:
    print "Usage: ./ban_user_and_stop_vessels.py <geni_username>"
    sys.exit(1)
  ban_user_and_remove_vessels(sys.argv[1])



"""
<Program>
  nodemanager.py

<Started>
  28 June 2009

<Author>
  Justin Cappos
  Justin Samuel

<Purpose>
  This is the nodemanager api for seattlegeni.
  
  Unless you are are developing the backend server, all you will use this
  module for is to first call init_nodemanager() and then to call either of
  these two functions:
  
    * get_node_info()
    * get_vessel_resources()
  
  You will not be able to use any of the other functions in this module from
  most code because you will not have access to the private owner keys for the
  node. Only the backend should be accessing the private owner keys. That is,
  if you need to change the node state, use the backend api, not this api.
  
  Information for using the other functions:
  
  If you are developing the backend_daemon itself, then you have a legitimate
  reason to be calling the other functions in this module which can change the
  node state. To use those, you will first call init_nodemanager() and then do
  the following to use any of the functions that change node state:
  
    1. Call get_node_handle() to obtain a node handle that can be passed to
       the other functions in this module.
    2. Call any of the other functions in this module, passing the node handle
       as the first argument.
"""

import random
import traceback

from seattlegeni.common.util.assertions import *

from seattlegeni.common.exceptions import *

from seattlegeni.common.util.decorators import log_function_call_without_first_argument

# Let's keep a copy of the built-ins, as repyportability destroys them
import __builtin__
builtins = __builtin__.__dict__.copy()

from repyportability import *
add_dy_support(locals())

###### code to enable affixes #########
dy_import_module_symbols('affixstackinterface.r2py')

affix_obj = AffixStackInterface('(CoordinationAffix)')

old_openconnection = openconnection

def new_openconnection(destip, destport, localip, localport, timeout):
  # we may get a unicode string.  repy's network API only accepts
  # standard python strs.
  destip = str(destip)
  if destip.endswith('zenodotus.poly.edu'):
    return affix_obj.openconnection(destip, destport, localip, localport, timeout)
  else:
    return old_openconnection(destip, destport, localip, localport, timeout)

openconnection = new_openconnection

# overwrite the openconnection that is provided to modules we import
sys.modules['dylink_repy'].openconnection = openconnection

####### end code to enable affixes ######

dy_import_module_symbols('nmclient.r2py')
dy_import_module_symbols("listops.r2py")
dy_import_module_symbols("time.r2py")

# Restore built-ins so that django and other libraries don't complain
for i in builtins:
  __builtin__.__dict__[i] = builtins[i]



# The number of times to try to do a time update before init_nodemanager() will
# raise an exception.
MAX_TIME_UPDATE_ATTEMPTS = 5

# Ports to use for UDP listening when doing a time update.
TIME_UPDATE_POSSIBLE_PORTS = range(10000, 60001)





def init_nodemanager():
  """
    <Purpose>
      Initializes the nodemanager api. Must be called before other operations.
    <Arguments>
      None
    <Exceptions>
      TimeUpdateError
        If unable to update time after MAX_TIME_UPDATE_ATTEMPTS tries.
    <Side Effects>
      This function contacts a NTP server and gets the current time.
      This is needed for the crypto operations that we do later.
      This selects random ports to listen on from the range
      TIME_UPDATE_POSSIBLE_PORTS.
    <Returns>
      None
  """
  portlist = random.sample(TIME_UPDATE_POSSIBLE_PORTS, MAX_TIME_UPDATE_ATTEMPTS) 
  
  for localport in portlist:
    try:
      time_updatetime(localport)
      return
    except TimeError:
      error_message = traceback.format_exc()
  
  # We raise a TimeUpdateError which is a seattlegeni error rather than a
  # TimeError which is a repy error.
  raise TimeUpdateError("Failed to perform time_updatetime(): " + error_message)
    



def get_node_info(ip, port):
  """
  <Purpose>
    Query a nodemanager for information about it.
  <Arguments>
    ip
      The ip address of of the nodemanager.
    port
      The port the nodemanager is listening on.
  <Exceptions>
    NodemanagerCommunicationError
      If we cannot communicate with a nodemanager at the specified ip and port.
  <Side Effects>
    None
  <Returns>
    A dictionary as returned by nmclient_getvesseldict(). This is a dictionary
    that will have at least the following keys (slight difference from the the
    result of nmclient_getvesseldict(), as that doesn't promise the keys exist):
      version
      nodename
      nodekey
      vessels (a dict where the keys are the vessel names and the values are dict's)
        [firstvesselname]
          userkeys
          ownerkey
          ownerinfo
          status
          advertise
        [secondvesselname]
          ...
    Node that even though we promise the keys exist, when the value is a key it may
    be None rather than a key dictionary that rsa.r2py likes.
  """
  assert_str(ip)
  assert_int(port)
  
  try:
    # This can raise an NMClientException, but the handle won't be stored in
    # the nmclient module if it does so we don't have to clean it up.
    nmhandle = nmclient_createhandle(ip, port)

    # Be sure to clean up the handle.    
    try:
      nodeinfo = nmclient_getvesseldict(nmhandle)
    finally:
      nmclient_destroyhandle(nmhandle)
    
  except NMClientException:
    nodestr = str((ip, port))
    message = "Failed to communicate with node " + nodestr + ": "
    raise NodemanagerCommunicationError(message + traceback.format_exc())
  
  # It's not fun for the client code to have to remember to check whether keys
  # exist, so make sure they do.
  fullnodeinfo = {"version":"", "nodename":"", "nodekey":None, "vessels":{}}
  fullnodeinfo.update(nodeinfo)
  
  # The fullnodeinfo["vessels"] dict is now the same dict as that in nodeinfo.
  for vesselname in fullnodeinfo["vessels"]:
    fullvesselinfo = {"userkeys":[], "ownerkey":None, "ownerinfo":"",
                      "status":"", "advertise":False}
    fullvesselinfo.update(fullnodeinfo["vessels"][vesselname])
    fullnodeinfo["vessels"][vesselname] = fullvesselinfo
  
  return fullnodeinfo





def _get_vessel_usableports(resourcedata):
  """
  A helper function for get_vessel_resources().
  Finds the list of ports where the resource contains both connport and messport.
  """
  # I think this code could stand to be in nmclient.r2py.

  connports = []
  messports = []

  for line in resourcedata.split('\n'):

    if line.startswith('resource'):
      # Ignore the word "resource" and any comments at the end.
      (resourcetype, value) = line.split()[1:3]
      if resourcetype == 'connport':
        # We do int(float(x)) because value might be a string '13253.0'
        connports.append(int(float(value)))
      if resourcetype == 'messport':
        messports.append(int(float(value)))

  return listops_intersect(connports, messports)





def get_vessel_resources(ip, port, vesselname):
  """
  <Purpose>
    Query a nodemanager for information about a vessel's resources. Currently
    only obtains information about usableports, but can be expanded to include
    more information as needed.
  <Arguments>
    ip
      The ip address of of the nodemanager.
    port
      The port the nodemanager is listening on.
    vessel
      The vessel whose resource information we want.
  <Exceptions>
    NodemanagerCommunicationError
      If we cannot communicate with a nodemanager at the specified ip and port.
  <Side Effects>
    None
  <Returns>
    A dictionary that has the following keys:
      usableports -- the value of this key is a list of ports that the vessel
                     has available as both a connport and a messport.
  """
  assert_str(ip)
  assert_int(port)
  assert_str(vesselname)
  
  resourcesdict = {}
  
  try:
    # This can raise an NMClientException, but the handle won't be stored in
    # the nmclient module if it does so we don't have to clean it up.
    nmhandle = nmclient_createhandle(ip, port)
    
    # Be sure to clean up the handle.
    try:
      resourcedata = nmclient_rawsay(nmhandle, "GetVesselResources", vesselname)  
    finally:
      nmclient_destroyhandle(nmhandle)
    
    resourcesdict["usableports"] = _get_vessel_usableports(resourcedata)
    
    
  except NMClientException:
    nodestr = str((ip, port))
    message = "Failed to communicate with node " + nodestr + ": "
    raise NodemanagerCommunicationError(message + traceback.format_exc())
  
  return resourcesdict





def get_node_handle(nodeid, ip, port, pubkeystring, privkeystring):
  """
  <Purpose>
    Obtain a node handle that can be used with the node-state-changing
    functions in this module.
  <Arguments>
    nodeid
      The node's id (a.k.a. "nodekey").
    ip
      The ip address of of the nodemanager.
    port
      The port the nodemanager is listening on (an int or long).
    pubkeystring
      The public key string of the ownerkey that will be used for any signed
      nodemanager communication.
    privkeystring
      The private key string of the ownerkey that will be used for any signed
      nodemanager communication.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    An opaque object that can be passed as the node handle to other functions
    in this module.
  """
  assert_str(nodeid)
  assert_str(ip)
  assert_int(port)
  assert_str(pubkeystring)
  assert_str(privkeystring)
  
  return (privkeystring, (nodeid, ip, port, pubkeystring))





def change_users(nodehandle, vesselname, userkeylist):
  """
  <Purpose>
    Perform a ChangeUsers call on a vessel.
  <Arguments>
    nodehandle
      A node handle obtained through a call to get_node_handle().
    vesselname
      The name of the vessel.
    userkeylist
      A list of public key strings which are the user keys to be set on the
      vessel.
  <Exceptions>
    NodemanagerCommunicationError
      If we cannot communicate with a nodemanager or the request fails.
  <Side Effects>
    The user key list of the vessel has been replaced.
  <Returns>
    None
  """
  assert_str(vesselname)
  assert_list_of_str(userkeylist)
  
  _do_signed_call(nodehandle[0], nodehandle[1], 'ChangeUsers', vesselname, '|'.join(userkeylist))





def reset_vessel(nodehandle, vesselname):
  """
  <Purpose>
    Perform a ResetVessel call on a vessel.
  <Arguments>
    nodehandle
      A node handle obtained through a call to get_node_handle().
    vesselname
      The name of the vessel.
  <Exceptions>
    NodemanagerCommunicationError
      If we cannot communicate with a nodemanager or the request fails.
  <Side Effects>
    The vessel has been reset.
  <Returns>
    None
  """
  assert_str(vesselname)
  
  _do_signed_call(nodehandle[0], nodehandle[1], 'ResetVessel', vesselname)





def change_owner(nodehandle, vesselname, ownerkey):
  """
  <Purpose>
    Perform a ChangeOwner call on a vessel.
  <Arguments>
    nodehandle
      A node handle obtained through a call to get_node_handle().
    vesselname
      The name of the vessel.
    ownerkey
      The public key to set as the owner key of the vessel.
  <Exceptions>
    NodemanagerCommunicationError
      If we cannot communicate with a nodemanager or the request fails.
  <Side Effects>
    The owner key of the vessel has been replaced. Future requests that modify
    this vessel will need to use this new owner key.
  <Returns>
    None
  """
  assert_str(vesselname)
  assert_str(ownerkey)
  
  _do_signed_call(nodehandle[0], nodehandle[1], 'ChangeOwner', vesselname, ownerkey)





def split_vessel(nodehandle, vesselname, desiredresourcedata):
  """
  <Purpose>
    Perform a SplitVessel call on a vessel.
  <Arguments>
    nodehandle
      A node handle obtained through a call to get_node_handle().
    vesselname
      The name of the vessel.
    desiredresourcedata
      A string of resourcedata that specifies the resources of a new vessel to
      create when splitting the existing vessel. This resourcedata has the
      format of a resources file.
  <Exceptions>
    NodemanagerCommunicationError
      If we cannot communicate with a nodemanager or the request fails.
  <Side Effects>
    The vesselname no longer exists. It has been split into two new vessels,
    one of which has the resources specified in the desiredresourcedata.
  <Returns>
    A tuple of the two new vessel names that resulted from the split. The
    first element of the tuple is the name of the vessel that has the
    leftover resources from the split. The second element of the tuple is
    the name of the vessel that has the exact resources specified in the
    desiredresourcedata.
  """
  assert_str(vesselname)
  assert_str(desiredresourcedata)
  
  splitvesselretval = _do_signed_call(nodehandle[0], nodehandle[1],
                                      'SplitVessel', vesselname, desiredresourcedata)
  
  # Get the new vessel names. The "left" vessel has the leftovers, the 
  # "right" is of the size requested.
  leftovervesselname, exactvesselname = splitvesselretval.split()

  return (leftovervesselname, exactvesselname)





def join_vessels(nodehandle, firstvesselname, secondvesselname):
  """
  <Purpose>
    Perform a JoinVessels call on two vessels.
  <Arguments>
    nodehandle
      A node handle obtained through a call to get_node_handle().
    firstvesselname
      The name of the first vessel to be joined.
    secondvesselname
      The name of the second vessel to be joined.
  <Exceptions>
    NodemanagerCommunicationError
      If we cannot communicate with a nodemanager or the request fails.
  <Side Effects>
    Neither firstvesselname nor secondvesselname exist. The have been joined
    together into a vessel that has a new name.
  <Returns>
    The name of the new vessel that has been created by joining together the
    two vessels.
  """
  assert_str(firstvesselname)
  assert_str(secondvesselname)
  
  combinedvesselname = _do_signed_call(nodehandle[0], nodehandle[1],
                                       'JoinVessels', firstvesselname, secondvesselname)

  return combinedvesselname





@log_function_call_without_first_argument
def _do_signed_call(privkeystring, nodeid_ip_port_pubkey_tuple, *callargs):
  """
    <Purpose>
      Performs an action that requires authentication on a remote node.
      
      The arguments are a little weird because our goal was to keep the rest
      of the code clean but have the private key be a separate, first argument
      to this function. The reason for this is so that we can use a logging
      decorator but not log the private key. We wouldn't have mangled a public
      function this way, but this is private and the nodehandle is opaque
      to client code.
    <Arguments>
      privkeystring:
        The private key used for authentication
      nodeid_ip_port_pubkey_tuple:
        nodeid:
          The node's identifier.
        ip:
          The node's IP address (a string)
        port:
          The port that the node manager is running on (an int)
        pubkeystring:
          The public key used for authentication
      *callargs:
        The arguments to give the node.   The first argument will usually be
        the call type (i.e. "ChangeUsers")
    <Exceptions>
      Exception / NMClientException are raised when the call fails.   
    <Side Effects>
      Whatever side effects the call has on the remote node.
    <Returns>
      None
  """
  (nodeid, ip, port, pubkeystring) = nodeid_ip_port_pubkey_tuple
  
  try:
    # This can raise an NMClientException, but the handle won't be stored in
    # the nmclient module if it does so we don't have to clean it up.
    nmhandle = nmclient_createhandle(ip, port)
  
    # Be sure to clean up the handle.    
    try:
      myhandleinfo = nmclient_get_handle_info(nmhandle)
      myhandleinfo['publickey'] = rsa_string_to_publickey(pubkeystring)
      myhandleinfo['privatekey'] = rsa_string_to_privatekey(privkeystring)
      nmclient_set_handle_info(nmhandle, myhandleinfo)
    
      return nmclient_signedsay(nmhandle, *callargs)
      
    finally:
      nmclient_destroyhandle(nmhandle)
    
  except NMClientException:
    nodestr = str((nodeid, ip, port))
    message = "NodeManager request failed with node " + nodestr + ": "
    raise NodemanagerCommunicationError(message + traceback.format_exc())
  



"""
<Program>
  lockserver_daemon.py

<Started>
  30 June 2009

<Author>
  Justin Samuel

<Purpose>
  This is the XML-RPC Lockserver that is used by various components of
  SeattleGeni.
 
  Two lock types are supported: 'user' and 'node'.
 
  Multiple of the same lock type ('user' or 'node') can be requested at
  the same time, but a single lock acquisition request cannot include
  requests for both types of locks.
 
  Clients first request a session identifier from the lockserver and
  then make requests for acquiring and releasing locks using that
  session identifier. It is the client's responsibility to release any
  lock they have acquired and to let the lockserver know when they
  are finished with their session (that is, when they have released
  all of their locks and do not plan to acquire more locks in the
  same session).
 
  Clients are denied lock acquisition requests as follows:
    * If the client's session already holds 'user' locks and the client is
      requesting 'user' locks.
    * If the client's session already holds 'node' locks.
    * If the client's session has any pending (unfulfilled) lock requests.

  That is, a client must (in a given session) request and obtain user locks
  before requesting node locks if the client intends to hold both kinds of
  locks at the same time.
 
  If clients hold both user locks and node locks, the locks can be released
  in any order. As per the restrictions on lock acquisition, though, the
  client won't be able to obtain additional node locks until they have
  released all of the node locks they hold. They will also not be able
  to obtain additional user locks until they have released all of their locks.


XML-RPC Interface:
 
StartSession()
  <Purpose>
    Obtain a new session identifier that can be used to acquire/release locks.
  <Arguments>
    None.
  <Exceptions>
    None.
  <Side Effects>
    The new session identifier is reserved within the lockserver.
  <Returns>
    Returns a string with a new session id. The value will be unique with
    respect to all session ids currently being tracked by the lockserver.
 
EndSession(session_id_str)
  <Purpose>
    Destroys the specified session.
  <Arguments>
    session_id_str: the session id
  <Exceptions>
    This will evoke an xmlrpclib.Fault exception on the client side if the session
    does not exist or if the session has locked resources or pending lock requests.
  <Side Effects>
    Removes all information about the specified session.
    Since an Exception will occur if the session holds any locks or has any
    pending lock requests, it is mostly just removing the record of the session
    identifier.
  <Returns>
    None.
        
AcquireLocks(session_id_str, lockdict)
  <Purpose>
    Obtain a lock on one or more lock names of a given type. This is a blocking
    request.
  <Arguments>
    session_id_str: the session id
    lockdict: a "lockdict" (see notes below)
  <Exceptions>
    This will evoke an xmlrpclib.Fault exception on the client side if the request
    is for locks the client shouldn't be requesting. See the notes above that give 
    the reasons clients will be denied lock acquisition.
  <Side Effects>
    Blocks until all of the locks specified in the lockdict are obtained for
    the specified session.
  <Returns>
    None.
 
ReleaseLocks(session_id_str, lockdict)
  <Purpose>
    Release one or more locks of one or more types.
  <Arguments>
    session_id_str: the session id
    lockdict: a "lockdict" (see notes below)
  <Exceptions>
    This will evoke an xmlrpclib.Fault exception on the client side if the session
    does not hold one or more of the locks that are listed in the lockdict.
  <Side Effects>
    Releases the locks previously held by the session. The released locks
    will be immediately obtained by other sessions waiting on the same
    lock.
  <Returns>
    None.
 
GetStatus()
  <Purpose>
    Obtains information about locks that are held, who holds them, and
    what locks are needed by any sessions that are waiting on locks.
    This allows observing the current state of the lockserver.
  <Arguments>
    None.
  <Exceptions>
    None.
  <Side Effects>
    None.
  <Returns>
    A dictionary with the keys, "heldlockdict",  "sessiondict", and
    "locktimelist" is returned. The values of these keys are most of the
    contents of the global variables by the same names used within the
    lockserver itself. It excludes the Event objects from the sessiondict
    data that is returned.

 
Details of the "lockdict" format:
  The lockdict format is a dictionary with two possible valid keys, 'user'
  and 'node'. If the key exists, its value must be a list of strings. Each
  of the strings specifies the name of a lock (whose type is either 'user'
  or 'node').
   
  Examples of valid lockdicts:
    {'user':['bob']}
    {'node':['123','xyz']}
    {'user':['bob'], 'node':['123','xyz']}
   
  The last one above, {'node': ['bob'], 'user': ['123','xyz']}, would not
  be valid for an AcquireLocks request because it lists two different
  types of locks (both 'user' and 'node'). It would, however, be valid for a
  ReleaseLocks request.
   
  Examples of invalid lockdicts:
    {}
    {'user':None}
    {'user':[123]}     # the name of the lock must be a string, '123'
    {'user':[]}        # an empty list of names is not valid
    {'x':['bob']}      # 'x' is not a valid locktype (only 'user' and 'node')
 
 
<Notes> 
  The lock type of 'global' has not been implemented as there is no clear need
  for it at this time and it would increase complexity.
 
  The lockserver does not perform special handling of requests that disconnect
  from the server during a lock acquisition request. The client's session will
  ultimately be granted the locks (assuming it was a valid request).
 
  This implementation does not allow an individual session to make additional
  lock acquisition requests while an existing lock acquisition request by the
  same session is unfulfilled. It doesn't seem like it would be correct
  behavior on behalf of a client to do so, so it's not supported. One case
  where this might be an issue is with parallelized work by the same client
  code using the same session id.
 
 
  General overview of the code: 
 
  The functions that can be called through the xmlrpc interface are the the 
  public functions in the class called LockserverPublicFunctions. Each of those 
  functions sanitizes user input, obtains a global datalock (of which there is 
  one in the entire module), and calls a helper function to do the actual work. 
  The helper function for any public xmlrpc function named MyFunction is of the
  format do_my_function. The global datalock is not used anywhere but
  in the methods of LockserverPublicFunctions. All other code that modifies
  global data assumes that the global datalock is held.
 
  Each session is only allowed to make one call to AcquireLocks at a time.
  This is enforced by disallowing AcquireLocks setting a boolean flag in
  the data for that session to indicate the session has a pending AcquireLocks
  request.
 
  The blocking of request threads when all locks are not yet available is
  handled through the use of threading.Event() objects. Each session has
  a single Event() object. A single Event per session is adequate because
  each session is only allowed to make one AcquireLocks call at a time.
  The AcquireLocks call will wait on the session's Event object until
  it has been set, which is the signal that all of the requested locks
  have been acquired for that session. The event can either be set/signaled
  as a result of the AcquireLocks call (if all of the locks are available
  at the time of the request) or by ReleaseLocks calls made by any session
  (if a lock that is released leaves the blocked AcquireLocks request
  with no more locks needing to be acquired).
 

<TODOs>
  TODO: Test to find out how many blocked threads can be supported.
  TODO: Test that a blocked lock request will not disconnect if it
        is blocked for a very long time.
  TODO: Write more integration tests (at least to test invalid requests).
  TODO: Write a simple benchmark script.
"""

import datetime

import time
import sys

# To send the admins emails when there's an unhandled exception.
import django.core.mail 

from seattlegeni.common.exceptions import *

from seattlegeni.common.util import log

from seattlegeni.website import settings

# Use threading.Lock directly instead of repy's getlock() to ease testing
# by not depending on repy. We also use threading.Event().
import threading

# To start the background monitor_held_lock_times() thread.
import thread

import traceback

# These are used to build a single-threaded XMLRPC server.
import SocketServer
import SimpleXMLRPCServer

# Used to generate session ids.
import random



# The port that we'll listen on.
LISTENPORT = 8010

# Session ids will be random numeric strings between these two values, inclusive.
MIN_SESSION_ID = 1000000
MAX_SESSION_ID = 9999999

# This is the amount of time the longest-held lock is allowed to be held before
# we consider there to be something very wrong and send notification emails to
# administrators.
MAX_EXPECTED_LOCK_HOLDING_TIMEDELTA = datetime.timedelta(minutes=5)

# Number of seconds to wait between checks of whether a lock has been held too
# long.
SECONDS_BETWEEN_LOCK_HOLDING_TIME_CHECKS = 30






# Whether the lockserver encountered an error. The server server thread will
# use this to decide whether to exit.
lockserver_had_error = False

# A mutex for access to all shared data.
datalock = threading.Lock()

# Format is not the same as the lockdict described in the module comments at the top:
#heldlockdict = {
#                "user"
#                  "bob" : {
#                    "queue" : [list_of_session_ids],
#                    "locked_by_session" : None
#                  },
#                  ... 
#                },        
#                "node"
#                  [same as "user", that is, each node has its own key and the
#                   value is a dict containing keys "queue" and "locked_by_session"]
#                },     
#}
# Note: This value is initialized by the call to init_globals()
heldlockdict = None

# Keeps track of which locks are held by which lockserver clients.
# The keys are lockserver session ids and the values are dictionaries of locks
# that are held by that session id.
# Format is not the same as the lockdict described in the module comments at the top:
#sessiondict = {
#                "abc123" : {
#                  "heldlocks" : {
#                    "user" : [list_of_user_name_strings],
#                    "node" : [list_of_node_name_strings]
#                  },
#                  "neededlocks" : lockdict only containing unfulfilled items,
#                  "acquirelocksproceedevent" : Event object used to block an AcquireLocks request until it is fulfilled,
#                  "acquirelocksinprogress" : boolean value to indicate whether an AcquireLocks request is in progress
#                }
#              }
# Note: This value is initialized by the call to init_globals()
sessiondict = None

# This is a list of tuples of the format ({locktype: lockname}, locktime) 
# containing an entry for every acquired lock. The locktime is a datetime
# object representing the time when the lock was acquired. The list is in
# order where the first item in the list is the longest-held lock and the
# last item is the shortest-held lock. The reason this information is not
# just kept in the heldlockdict is largely because we don't want to return
# the time in with the GetStatus call because the tests would have to be
# changed to expect a value there that is different with every run of the
# test. Also, all we really care about is the longest-held lock and most
# of the point here is to be able to detect when any lock has been held
# past a threshold that we consider reasonable. So, for that, we might
# as well keep track of the order explicitly as we know that information
# here in the lockserver daemon.
locktimelist = []





class ThreadedXMLRPCServer(SocketServer.ThreadingMixIn, SimpleXMLRPCServer.SimpleXMLRPCServer):
  """This is a threaded XMLRPC Server. """




  
class LockserverInvalidRequestError(Exception):
  """Indicates that an invalid request was made by the client."""





def _lockdict_contains_lock(lockdict, locktype, lockname):
  """
  Returns True if a lock of the specified locktype and lockname is in lockdict,
  otherwise returns False.
  """
  if locktype in lockdict:
    if lockname in lockdict[locktype]:
      return True
  return False





def _is_lockdict_empty(lockdict):
  """
  Returns True if there are no locks specified in the lockdict, otherwise
  returns False. Note that there must be at least one lock name defined
  for at least one lock locktype for there to be considered a lock in the
  lockdict.
  """
  for locktype in lockdict:
    if len(lockdict[locktype]) > 0:
      return False
  return True





def init_globals():
  """
  <Purpose>
    Prepares the global variables heldlockdict and sessiondict. They
    are set this way rather than directly when declared as this this method is
    needed for unit tests that work directly with the lockserver_daemon module
    rather than starting and stopping the lockserver and using xmlrpc. This
    method should never be called after the lockserver is running.
  <Arguments>
    None.
  <Exceptions>
    None.
  <Side Effects>
    Resets the heldlockdict and sessiondict global variables, thus clearing
    the state of the lockserver.
  <Returns>
    None.
  """
  
  global heldlockdict
  global sessiondict

  heldlockdict = {"user":{}, "node":{}}
  sessiondict = {}
  
  
  


def _generate_session_id():
  """
  Generates a session id that doesn't already exist in the sessiondict. The
  created session id is a numeric string between MIN_SESSION_ID and
  MAX_SESSION_ID. We create intentionally random session ids rather than
  assigning them sequentially to reduce the risk of clients accidentally using 
  a session id that really isn't theirs.
  """
  session_id = None
  while session_id is None or session_id in sessiondict:
    session_id = str(random.randint(MIN_SESSION_ID, MAX_SESSION_ID))
    
  return str(session_id)





def do_start_session():
  """
  <Purpose>
    This is the function that does the actual work for xmlrpc calls to
    StartSession. The caller of this function must hold the global datalock.
  <Arguments>
    None.
  <Exceptions>
    None.
  <Side Effects>
    The new session id is added to the global sessiondict.
  <Returns>
    The newly-created session id.
  """
  session_id = _generate_session_id()
  
  sessiondict[session_id] = {}
  
  # The Event object is to cause lock acquisition requests by this session to
  # block until the request has been fulfilled.
  sessiondict[session_id]["acquirelocksproceedevent"] = threading.Event()
  # Have the event be initially set. This is so that we can know that the
  # event is only unset (clear) if an AcquireLocks thread is blocked and
  # waiting for locks it needs to acquire. Though the code should work
  # fine even if we don't set this here, it's better to be able to conclusively
  # determine whether a session has a blocked AcquireLocks request by inspecting
  # this (making testing and debugging easier).
  sessiondict[session_id]["acquirelocksproceedevent"].set()
  
  # Indicate whether there is a current AcquireLocks call that has not been
  # fulfilled yet. We use this so that an AcquireLocks request thread can
  # indicate that it has proceeded past its event wait() call and thus the
  # same session is allowed to make further AcquireLocks requests.
  sessiondict[session_id]["acquirelocksinprogress"] = False

  # Create empty lockdicts with "user" and "node" keys for indicating the
  # locks the session holds and the locks the session is queued for.
  sessiondict[session_id]["heldlocks"] = {"user":[], "node":[]}
  sessiondict[session_id]["neededlocks"] = {"user":[], "node":[]}
  
  return session_id





def do_end_session(session_id):
  """
  <Purpose>
    This is the function that does the actual work for xmlrpc calls to
    EndSession. The caller of this function must hold the global datalock.
  <Arguments>
    session_id:
      The string that is the session id to be ended.
  <Exceptions>
    LockserverInvalidRequestError is raised if the specified session does
    not exist, if it still holds locks, or if it has pending, unfulfilled
    lock acquisition requests.
  <Side Effects>
    If no exception occurs, the specified session is removed from the
    global sessiondict.
  <Returns>
    None.
  """
  # Raises an exception if the session id doesn't exist.
  _assert_valid_session(session_id)
    
  if not _is_lockdict_empty(sessiondict[session_id]["heldlocks"]):
    raise LockserverInvalidRequestError("Cannot end session: this session still holds locks.")
    
  if not _is_lockdict_empty(sessiondict[session_id]["neededlocks"]):
    raise LockserverInvalidRequestError("Cannot end session: this session has pending queued lock requests.")
    
  # Get rid of the session.
  del sessiondict[session_id]
    




def do_acquire_locks(session_id, requested_acquire_lockdict):
  """
  <Purpose>
    This is the function that does the actual work for xmlrpc calls to
    AcquireLocks. Other than for testing, this should only be called by the
    AcquireLocks function registered with the xmlrpc server.
    The caller of this function must hold the global datalock.
  <Arguments>
    session_id:
      The string that is the session id under which the locks should be acquired.
    requested_acquire_lockdict:
      The lockdict that contains the locks to be acquired.
  <Exceptions>
    LockserverInvalidRequestError is raised if the specified session is
    invalid or the requested locks are invalid (including whether they are
    invalid based on locks the session holds or previous lock acquisitions
    requests that have not been fulfilled yet).
  <Side Effects>
    Each specified lock is either immediately granted (if no other session
    holds it) or this session is added to a queue for that lock. The Event
    that belongs the specified session is reset (cleared) to ensure the
    server's thread handling this request blocks. As soon as all of the
    requested locks are held, the Event is used to allow the server's thread
    to proceed (and thus respond to the client).
  <Returns>
    None.
  """
  # Raises an exception if the session id doesn't exist.
  _assert_valid_session(session_id)
  
  # Raises an exception if the lockdict format is invalid.
  _assert_valid_lockdict(requested_acquire_lockdict)
  
  # Raises an exception if the requested locks are invalid, including if they
  # conflict with ones held by the same session.
  _assert_valid_locks_for_acquire(session_id, requested_acquire_lockdict)
  
  for locktype in requested_acquire_lockdict:
    for lockname in requested_acquire_lockdict[locktype]:
      _acquire_individual_lock(session_id, locktype, lockname)
      
  # Check if the request got all of the locks it asked for in order to
  # determine whether the request thread should block.
  if _is_lockdict_empty(sessiondict[session_id]["neededlocks"]):
    # It did get all of the locks it asked for, so the request thread shouldn't block.
    sessiondict[session_id]["acquirelocksproceedevent"].set()
  else:
    # It did not get all of the locks it asked for, so the request thread should block.
    sessiondict[session_id]["acquirelocksproceedevent"].clear()





def _acquire_individual_lock(session_id, locktype, lockname):
  """
  <Purpose>
    This is called by do_acquire_locks for each lock to be acquired. This will
    either mark the lock as being held by the specified session (if the lock
    is not already held) or will add this session the lock's queue (if the
    lock is already held).
  <Arguments>
    session_id:
      The string that is the session id under which the locks should be acquired.
    locktype:
      The locktype of lock, either 'user' or 'node'.
    lockname:
      The lockname of the lock (a string).
  <Exceptions>
    None.
  <Side Effects>
    Modifies the global heldlockdict and the global sessiondict based on
    whether the lock is immediately given to the session or whether the
    session has been put in the lock's queue.
  <Returns>
    None.
  """
  # Create the lock info if it didn't already exist in the global heldlockdict.
  if not lockname in heldlockdict[locktype]:
    heldlockdict[locktype][lockname] = {"queue":[], "locked_by_session":None}

  heldlockinfo = heldlockdict[locktype][lockname]
  
  if heldlockinfo["locked_by_session"] is None:
    # Nobody holds this lock, so give it to this session.
    heldlockinfo["locked_by_session"] = session_id
    
    # Record in the sessiondict that the session holds this lock.
    sessiondict[session_id]["heldlocks"][locktype].append(lockname)
    
    # Append to the locktimelist indicating when this lock was acquired.
    locktimelist.append(({locktype: lockname}, datetime.datetime.now()))
    
  else:
    # This lock is already held, so add the session to this lock's queue.
    heldlockinfo["queue"].append(session_id)
    
    # Record in the sessiondict that this session is waiting on this lock.
    sessiondict[session_id]["neededlocks"][locktype].append(lockname)





def _assert_valid_locks_for_acquire(session_id, requested_acquire_lockdict):
  """
  <Purpose>
    Ensures that the locks specified in requested_acquire_lockdict are locks
    that the given session should legitimately be requesting. That is,
    ensures they aren't locks the session already holds, ensures that the
    session should even be requesting locks at all (e.g. do they have a pending
    lock request?), checks if they are requesting a 'user' lock even though
    they already hold a 'node' lock, etc.
  <Arguments>
    session_id:
      The string that is the session id under which the locks should be acquired.
    requested_acquire_lockdict:
      The lockdict that contains the locks to be acquired.
  <Exceptions>
    Raises LockserverInvalidRequestError if requested_acquire_lockdict contains
    locks that would be invalid for this session to request acquisition of.
  <Side Effects>
    None.
  <Returns>
    None.
  """
  # Check if the session is requesting locks of both type "user" and type "node".
  if len(requested_acquire_lockdict.keys()) != 1:
    # Raise an error that will be returned over xmlrpc.
    message = "Requested acquisition of locks of multiple types (you can only request locks of a single type at a time)."
    _raise_lock_request_error(session_id, requested_acquire_lockdict, message)
  
  # Locks of the same locktype as those already held cannot be requested.
  for locktype in requested_acquire_lockdict:
    # Check if the session already holds locks of this locktype.
    if len(sessiondict[session_id]["heldlocks"][locktype]) > 0:
      # Raise an error that will be returned over xmlrpc.
      message = "Requested acquisition of locks of same locktype ('" + locktype + "') as those already held by this session."
      _raise_lock_request_error(session_id, requested_acquire_lockdict, message)

  # User locks cannot be requested when a node lock is already held.
  if "user" in requested_acquire_lockdict:
    # Check if the session already holds locks of the 'node' locktype.
    if len(sessiondict[session_id]["heldlocks"]["node"]) > 0:
      # Raise an error that will be returned over xmlrpc.
      message = "Requested acquisition of user lock when node locks already held by this session."
      _raise_lock_request_error(session_id, requested_acquire_lockdict, message)

     
    
    

def do_release_locks(session_id, requested_release_lockdict):
  """
  <Purpose>
    This is the function that does the actual work for xmlrpc calls to
    ReleaseLocks. Other than for testing, this should only be called by the
    ReleaseLocks function registered with the xmlrpc server.
    The caller of this function must hold the global datalock.
  <Arguments>
    session_id:
      The string that is the session id under which the locks should be released.
    requested_release_lockdict:
      The lockdict that contains the locks to be released.
  <Exceptions>
    LockserverInvalidRequestError is raised if the specified session is
    invalid or if any of the locks requested to be released are not
    held by the session.
  <Side Effects>
    Each specified lock is immediately released. For any released lock, if
    there is a session waiting (queued) for the lock, the next queued session
    is given the lock (and the waiting thread unblocked if they do not
    require any more locks).
  <Returns>
    None.
  """
  # Raises an exception if the session id doesn't exist.
  _assert_valid_session(session_id)
  
  # Raises an exception if the lockdict format is invalid.
  _assert_valid_lockdict(requested_release_lockdict)
  
  # Raises an exception if the requested locks for release are invalid, including
  # if they are not all held by this session.
  _assert_valid_locks_for_release(session_id, requested_release_lockdict)
  
  for locktype in requested_release_lockdict:
    for lockname in requested_release_lockdict[locktype]:
      _release_individual_lock(session_id, locktype, lockname)





def _release_individual_lock(session_id, locktype, lockname):
  """
  <Purpose>
    This is called by do_release_locks for each lock to be released. This will
    mark the lock as not being held by the specified session and will take care
    of giving released locks to queued requests.
  <Arguments>
    session_id:
      The string that is the session id under which the locks should be released.
    locktype:
      The locktype of lock, either 'user' or 'node'.
    lockname:
      The lockname of the lock (a string).
  <Exceptions>
    None.
  <Side Effects>
    Modifies the global heldlockdict and the global sessiondict to indicate
    that the specified lock is no longer held by the session as well as to
    grant the lock to the next session in the lock's queue which is waiting
    for it, if any. If there was a session queued for this lock, after
    giving the lock to that session this function will check if the new
    lock holder is waiting on any more locks. If not, the server thread
    for the queued request will be unblocked.
  <Returns>
    None.
  """
  heldlockinfo = heldlockdict[locktype][lockname]
  
  # Regardless of whether there are queued sessions waiting for this lock,
  # it is removed from the list of locks this session holds.
  sessiondict[session_id]["heldlocks"][locktype].remove(lockname)

  # Remove this lock from the locktimelist.
  for locktimeitem in locktimelist:
    if locktimeitem[0] == {locktype: lockname}:
      log.info("Lock " + str({locktype: lockname}) + " was held for " + 
               str(datetime.datetime.now() - locktimeitem[1]))
      locktimelist.remove(locktimeitem)
      break
  
  if len(heldlockinfo["queue"]) > 0:
    # Set the lock as held by the next queued session_id.
    new_lock_holder = heldlockinfo["queue"].pop(0)
    heldlockinfo["locked_by_session"] = new_lock_holder
    
    # Update the sessiondict to change this lock from a needed lock to a held lock.
    sessiondict[new_lock_holder]["heldlocks"][locktype].append(lockname)
    sessiondict[new_lock_holder]["neededlocks"][locktype].remove(lockname)
    
    # Append to the locktimelist indicating when this lock was acquired.
    locktimelist.append(({locktype: lockname}, datetime.datetime.now()))
    
    # If the session  now holding the lock isn't waiting on any more locks,
    # unblock the session's current AcquireLocks request thread.
    if _is_lockdict_empty(sessiondict[new_lock_holder]["neededlocks"]):
      sessiondict[new_lock_holder]["acquirelocksproceedevent"].set()
    
  else:
    # There are no sessions waiting on this lock, so the lock is now held by nobody.
    heldlockinfo["locked_by_session"] = None
    
    
    
    
    
def _assert_valid_locks_for_release(session_id, requested_release_lockdict):
  """
  <Purpose>
    Ensures that the locks specified in requested_acquire_lockdict are locks
    that the given session should legitimately be trying to release. That is,
    ensures the session holds all of the locks specified in the lockdict.
  <Arguments>
    session_id:
      The string that is the session id under which the locks should be released.
    requested_release_lockdict:
      The lockdict that contains the locks to be released.
  <Exceptions>
    Raises LockserverInvalidRequestError if requested_release_lockdict contains
    any locks that the session does not hold.
  <Side Effects>
    None.
  <Returns>
    None.
  """
  sessionheldlockdict = sessiondict[session_id]["heldlocks"]
  
  # Go through the locks being requested for release and if any of them aren't
  # currently held by this session, raise an error.
  for locktype in requested_release_lockdict:
    for lockname in requested_release_lockdict[locktype]:
      if not _lockdict_contains_lock(sessionheldlockdict, locktype, lockname):
        # Raise an error that will be returned over xmlrpc.
        message = "Attempted to release locks not held by this session."
        _raise_lock_request_error(session_id, requested_release_lockdict, message)
    
  # Should we make sure the client isn't releasing a user lock while they still
  # hold node locks? There may not be any deadlock risk that results from
  # this, but it is likely incorrect behavior. Not forbidding it for now.




def do_get_status():
  """
  <Purpose>
    This is the function that does the actual work for xmlrpc calls to
    GetStatus. Other than for testing, this should only be called by the
    GetStatus function registered with the xmlrpc server.
    The caller of this function must hold the global datalock.
  <Arguments>
    None.
  <Exceptions>
    None.
  <Side Effects>
    None.
  <Returns>
    A dictionary with the keys, "heldlockdict",  "sessiondict", and
    "locktimelist" is returned. The values of these keys are most of the
    contents of the global variables by the same names used within the
    lockserver itself. It excludes the Event objects from the sessiondict
    data that is returned.
  """

  # We create a sessiondict that 
  cleansessiondict = {}
  for session_id in sessiondict:
    cleansessiondict[session_id] = {}
    cleansessiondict[session_id]["heldlocks"] = sessiondict[session_id]["heldlocks"]
    cleansessiondict[session_id]["neededlocks"] = sessiondict[session_id]["neededlocks"]
    
    # We include a acquirelocksproceedeventset value in the status info rather than the
    # boolean acquirelocksinprogress value because we want to be able to test without
    # running the xmlrpc server. That is, the unit tests that make direct
    # calls to do_acquire_locks() and do_release_locks() won't see a useful value 
    # for acquirelocksinprogress because that is only set directly in AcquireLocks.
    # However, the event that is waited on is set and cleared in in the
    # do_acquire_locks() and do_release_locks() functions, so it is useful for
    # unit testing (the code was intentionally organized to make that testable).
    # Note: python <2.6 only supports isSet(), not is_set().
    acquirelocksproceedeventset = sessiondict[session_id]["acquirelocksproceedevent"].isSet()
    cleansessiondict[session_id]["acquirelocksproceedeventset"] = acquirelocksproceedeventset
  
  status = {}
  status["heldlockdict"] = heldlockdict
  status["sessiondict"] = cleansessiondict
  status["locktimelist"] = locktimelist
  return status
    
    



def _assert_number_of_arguments(functionname, args, exact_number):
  """
  <Purpose>
    Ensure that an args tuple which one of the public xmlrpc functions was
    called with has an expected number of arguments.
  <Arguments>
    functionname:
      The name of the function whose number of arguments are being checked.
      This is just for logging in the case that the arguments don't match.
    args:
      A tuple of arguments (received by the other function through *args).
    exact_number:
      The exact number of arguments that must be in the args tuple.
  <Exceptions>
    Raises LockserverInvalidRequestError if args does not contain exact_number
    items.
  <Side Effects>
    None.
  <Returns>
    None.
  """
  if len(args) != exact_number:
    message = "Invalid number of arguments to function " + functionname + ". "
    message += "Expected " + str(exact_number) + ", received " + str(len(args)) + "."
    raise LockserverInvalidRequestError(message)





def _assert_valid_session(session_id):
  """
  <Purpose>
    Ensure that a session id received by a call to a public xmlrpc function
    is a valid session id (not just in format, but that the session was
    started and has not already been ended).
  <Arguments>
    session_id:
      The session id string to check.
  <Exceptions>
    Raises LockserverInvalidRequestError if session_id is not an existing
    session id.
  <Side Effects>
    None.
  <Returns>
    None.
  """
  if not isinstance(session_id, str):
    raise LockserverInvalidRequestError("Invalid session id (must be a string). You provided a " + str(type(session_id)) + " which was " + str(session_id))
  if session_id not in sessiondict:
    raise LockserverInvalidRequestError("Invalid session id (the specified session id doesn't exist).")





def _assert_valid_lockdict(lockdict):
  """
  <Purpose>
    Ensures that the specified lockdict (which comes from user input) is of an
    acceptable format.
  <Arguments>
    lockdict:
      The lockdict to check.
  <Exceptions>
    Raises LockserverInvalidRequestError the specified lockdict is illegal for
    any reason (e.g. isn't a dictionary, specifies lock types that are invalid,
    etc.)
  <Side Effects>
    None.
  <Returns>
    None.
  """
  if not isinstance(lockdict, dict):
    raise LockserverInvalidRequestError("Invalid lockdict (lockdict must be a dict).")
  
  if len(lockdict.keys()) == 0:
    raise LockserverInvalidRequestError("Invalid lockdict (must specify at least one locktype of lock).")

  for locktype in lockdict:
    if not isinstance(lockdict[locktype], list):
      raise LockserverInvalidRequestError("Invalid lockdict (each key's value must be a list of locknames).")
    
    if locktype not in ["user", "node"]:
      # Even though xmlrpclib doesn't allow non-strings for dict keys, python does.
      # So, to be proper/safe, we cast 'locktype' to a string.
      raise LockserverInvalidRequestError("Invalid lockdict (lock locktype '" + str(locktype) + "' does not exist).")
    
    if len(lockdict[locktype]) == 0:
      raise LockserverInvalidRequestError("Invalid lockdict (no lock names specified for lock locktype '" + locktype + "').")

    for lockname in lockdict[locktype]:
      if not isinstance(lockname, str):
        raise LockserverInvalidRequestError("Invalid lockdict (all items in a list of locknames must be str's).")
      if len(lockname) == 0:
        raise LockserverInvalidRequestError("Invalid lockdict (lock names cannot be empty strings).")
      if len(lockdict[locktype]) != len(set(lockdict[locktype])):
        raise LockserverInvalidRequestError("Invalid lockdict (all items in a list of locknames must unique in that list).")
      




def _raise_lock_request_error(session_id, lockdict_in_request, message):
  """
  Raises a LockserverInvalidRequestError with a description that includes
  the given session_id, lockdict_in_request, and message, as well as
  the session's held locks and pending locks.
  """
  requested_lock_str = str(lockdict_in_request)
  
  sessionheldlockdict = sessiondict[session_id]["heldlocks"]
  held_locks_str = str(sessionheldlockdict)
  
  sessionneededlockdict = sessiondict[session_id]["neededlocks"]
  needed_locks_str = str(sessionneededlockdict)
  
  info_str = "Session id: " + session_id + ". "
  info_str += "Locks in request: " + requested_lock_str + ". "
  info_str += "Held locks: " + held_locks_str + ". "
  info_str += "Pending locks: " + needed_locks_str + "."
  
  raise LockserverInvalidRequestError("Illegal request: " + message + " [" + info_str + "]")





class LockserverPublicFunctions(object):
  """
  All public functions of this class are automatically exposed as part of the
  xmlrpc interface.
  """
  
  def _dispatch(self, method, args):
    """
    We provide a _dispatch function (which SimpleXMLRPCServer looks for and
    uses) so that we can log exceptions due to our programming errors within
    thelockserver as well to detect incorrect usage by clients. When an
    internal lockserver error is detected, this method will signal to the main
    server thread to shutdown.
    """
    global lockserver_had_error
      
    try:
      # Get the requested function (making sure it exists).
      try:
        func = getattr(self, method)
      except AttributeError:
        raise LockserverInvalidRequestError("The requested method '" + method + "' doesn't exist.")
      
      # Call the requested function.
      return func(*args)
    
    except LockserverInvalidRequestError:
      log.error("The lockserver was used incorrectly: " + traceback.format_exc())
      raise
    
    except:
      # We assume all other exceptions are bugs in the lockserver.
      # If there is a bug in the lockserver, that's really bad. We terminate the
      # lockserver in this case rather than risk incorrect locking behavior.
      
      # This will tell the server thread to exit.
      lockserver_had_error = True

      message = "The lockserver had an internal error and is exiting." + traceback.format_exc()
      log.critical(message)

      # Send an email to the addresses listed in settings.ADMINS
      if not settings.DEBUG:
        subject = "Critical SeattleGeni lockserver error"
        django.core.mail.mail_admins(subject, message)
      
      # This request will likely end up seeing an xmlrpclib.ProtocolError due to the
      # shutdown, regardless of this exception.
      raise



  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  def StartSession(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('StartSession', args, 0)
    
    datalock.acquire()
    try:
      session_id = do_start_session()
      
      log.info("[session_id: " + session_id + "] StartSession called.")
      
      return session_id
    
    finally:
      datalock.release()
      
      
      
  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  def EndSession(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('EndSession', args, 1)
    # avoid python magic comma needed to write this as "(session_id,) = args"
    session_id = args[0]
    
    datalock.acquire()
    try:
      # Ensure it's a string before printing it like one.
      _assert_valid_session(session_id)
      
      log.info("[session_id: " + session_id + "] EndSession called.")
      
      do_end_session(session_id)
      
    finally:
      datalock.release()
      


  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  def AcquireLocks(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('AcquireLocks', args, 2)
    (session_id, request_acquire_lockdict) = args
    
    datalock.acquire()
    try:
      # Ensure it's a string before printing it like one.
      _assert_valid_session(session_id)
      
      log.info("[session_id: " + session_id + "] AcquireLocks called for locks " + str(request_acquire_lockdict))
    
      # Check if this session has an outstanding AcquireLocks request. Clients
      # should not be making concurrent AcquireLocks requests.
      if sessiondict[session_id]["acquirelocksinprogress"]:
        message = "[session_id: " + session_id + "] AcquireLocks called while an earlier AcquireLocks call has not been completed."
        raise LockserverInvalidRequestError(message)
      
      do_acquire_locks(session_id, request_acquire_lockdict)
      
      # Indicate that there is a running AcquireLocks request for this session
      # so that future AcquireLocks requests will be denied until this one is
      # fulfilled. If the call to do_acquire_locks raised an exception, this
      # will not get set.
      sessiondict[session_id]["acquirelocksinprogress"] = True
      
    finally:
      datalock.release()
    
    # Wait for our event flag to signal that we have acquired the locks.
    # This is what causes the request thread to block until it is fulfilled.
    # If the event is not set at this point (causing this thread to block),
    # then it will be set by calls to ReleaseLocks. If a call to ReleaseLocks
    # signals this event between the time we released the global datalock and
    # when we get to this wait() line, that's fine.
    sessiondict[session_id]["acquirelocksproceedevent"].wait()
    
    # Indicate that we've made it past our wait() call, meaning that future
    # calls to do_acquire_locks() can be allowed again. We do not need to hold
    # the global datalock to set this as it can only be set to True in the
    # critical section above which first ensures that the value is not already
    # true.
    sessiondict[session_id]["acquirelocksinprogress"] = False
    
    log.info("[session_id: " + session_id + "] AcquireLocks fulfilled request for locks " + str(request_acquire_lockdict))
    


  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  def ReleaseLocks(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('ReleaseLocks', args, 2)
    (session_id, request_release_lockdict) = args
  
    datalock.acquire()
    try:
      # Ensure it's a string before printing it like one.
      _assert_valid_session(session_id)
      
      log.info("[session_id: " + session_id + "] ReleaseLocks called for locks " + str(request_release_lockdict))
      
      do_release_locks(session_id, request_release_lockdict)
      
    finally:
      datalock.release()
  
  
  
  # Using @staticmethod makes it so that 'self' doesn't get passed in as the first arg.
  @staticmethod
  def GetStatus(*args):
    """
    This is a public function of the XMLRPC server. See the module comments at
    the top of the file for a description of how it is used.
    """
    _assert_number_of_arguments('GetStatus', args, 0)
    
    datalock.acquire()
    try:
      log.info("GetStatus called.")
      return do_get_status()
    finally:
      datalock.release()





def monitor_held_lock_times():
  """
  Periodically checks whether there are locks that have been held too long.
  When there is a lock that has been held too long, logs it and also sends
  an email if settings.DEBUG is False.
  
  This function gets started in its own thread.
  """
  
  log.info("[monitor_held_lock_times] thread started.")

  # Run forever.
  while True:
    
    try:
      
      # Wait a bit between checks.
      time.sleep(SECONDS_BETWEEN_LOCK_HOLDING_TIME_CHECKS)
      
      # Grab the datalock and get the oldest held lock, if there are any.
      datalock.acquire()
      try:
        if len(locktimelist) == 0:
          # No locks are held.
          continue
        
        oldestlock = locktimelist[0]
        
      finally:
        datalock.release()
        
      held_timedelta = datetime.datetime.now() - oldestlock[1]
      
      # Check if the oldest lock has been held too long.
      if held_timedelta > MAX_EXPECTED_LOCK_HOLDING_TIMEDELTA:
        message = "Lockserver lock " + str(oldestlock[0])
        message += " has been held since " + str(oldestlock[1])
        message += " (timedelta: " + str(held_timedelta) + ")"
        # Raise an exception which will cause an email to be sent from the
        # except clause below.
        raise InternalError(message)
        
    # Catch all exceptions so that the monitor thread will never die.
    except:
      message = "[monitor_held_lock_times] Something very bad happened: " + traceback.format_exc()
      log.critical(message)
      
      # Send an email to the addresses listed in settings.ADMINS
      if not settings.DEBUG:
        subject = "Critical SeattleGeni lockserver error"
        django.core.mail.mail_admins(subject, message)
        
        # Sleep for 30 minutes to make sure we don't flood the admins with error
        # report emails.
        time.sleep(60 * 30)





def main():

  # Initialize global variables.
  init_globals()

  # Register the XMLRPCServer. Use allow_none to allow allow the python None value.
  server = ThreadedXMLRPCServer(("127.0.0.1", LISTENPORT), allow_none=True)

  log.info("Listening on port " + str(LISTENPORT) + ".")
  
  # Start the background thread that watches for locks being held too long.
  thread.start_new_thread(monitor_held_lock_times, ())

  server.register_instance(LockserverPublicFunctions()) 
  while True:
    server.handle_request()
    # Shutdown the lockserver if there was an internal error.
    # This doesn't actually get detected until another request has been
    # made, as the main server thread is often already blocked in the
    # next handle_request() call when this this value get set.
    if lockserver_had_error:
      sys.exit(1)



if __name__ == '__main__':
  try:
    main()
  except KeyboardInterrupt:
    log.info("Exiting on KeyboardInterrupt.")
    sys.exit(0)



# Modules to make available for convenience so the names are available in the
# ipython shell. Just a lazy way to not have to execute these lines individually
# in a new shell.
from seattlegeni.common.api import maindb
from seattlegeni.common.exceptions import *
from seattlegeni.website.control import interface

import random

def testfailed(reason):
  raise Exception("Test failed: " + reason)

username = 'testuser' + str(random.randint(0, 1000000))

# Create a user.
geniuser = interface.register_user(username, 'mypass', 'joe@example.com', 'myaffiliation')

# Get the user.
geniuser = interface.get_user_with_password(username, 'mypass')

# Make sure the private key was created because we didn't provide a pubkey when
# creating the user.
privkey = interface.get_private_key(geniuser)

if privkey is None:
  testfailed("private key wasn't created")

# Delete the private key and make sure it gets deleted.
interface.delete_private_key(geniuser)

privkey = interface.get_private_key(geniuser)
if privkey is not None:
  testfailed("private key wasn't deleted: " + str(privkey))

# Make sure they don't have any donations initially.
donations = interface.get_donations(geniuser)
if len(donations) > 0:
  testfailed("The user shouldn't have any donations.")

INITIAL_DONATION_COUNT = 6

# Create a few nodes and donations by this user.
for i in range(INITIAL_DONATION_COUNT):
  node_identifier = 'testnode' + username + str(i)
  extra_vessel_name = 'v2'
  ip = '127.0.' + str(i) + '.0'
  node = maindb.create_node(node_identifier, ip, 1234, '0.1a', True, 'the owner pubkey', extra_vessel_name)
  maindb.create_donation(node, geniuser, 'some resource description text')
  # Create the vessels on these nodes.
  name = 'testvessel' + username + str(i)
  vessel = maindb.create_vessel(node, name)
  # Set the vessel ports.
  ports = range(1000, 1010)
  ports.append(geniuser.usable_vessel_port)
  maindb.set_vessel_ports(vessel, ports)

donations = interface.get_donations(geniuser)
if len(donations) != INITIAL_DONATION_COUNT:
  testfailed("The user's donation count doesn't equal their number of donations.")

# Make sure they don't have any acquired vessels initially.
acquired_vessels = interface.get_acquired_vessels(geniuser)
if len(acquired_vessels) > 0:
  testfailed("The user shouldn't have any acquired vessels.")

# Try to acquire vessels when there are none that match what they request.
# At this point, we expect the user to have 70 vessel credits (default 10
# plus 6 * 10 for donations).
try:
  interface.acquire_vessels(geniuser, 50, 'wan')
except UnableToAcquireResourcesError: 
  pass # This is what we expected.
else:
  testfailed("Didn't throw expected UnableToAcquireResourcesError")

# Try to acquire more vessels than the user should be allowed to.
try:
  interface.acquire_vessels(geniuser, 100, 'wan')
except InsufficientUserResourcesError:
  pass # This is what we expected.
else:
  testfailed("Didn't throw expected InsufficientUserResourcesError")

# Acquire INITIAL_DONATION_COUNT wan vessels. There should be at least this
# many because we just made them.
acquired_vessels = interface.acquire_vessels(geniuser, INITIAL_DONATION_COUNT, 'wan')

if len(acquired_vessels) != INITIAL_DONATION_COUNT:
  testfailed("Didn't acquire the number of vessels expected.")

# Release one vessel.
vessel_to_release = acquired_vessels[0]
interface.release_vessels(geniuser, [vessel_to_release])

acquired_vessels = interface.get_acquired_vessels(geniuser)
if len(acquired_vessels) != INITIAL_DONATION_COUNT - 1:
  testfailed("Wrong number of acquired vessels after releasing one.")
  
# Try to release the same one again.
try:
  interface.release_vessels(geniuser, [vessel_to_release])
except InvalidRequestError:
  pass # This is what we expected.
else:
  testfailed("Didn't throw expected InvalidRequestError")

# Release the rest of the user's vessels.
interface.release_all_vessels(geniuser)

acquired_vessels = interface.get_acquired_vessels(geniuser)
if len(acquired_vessels) != 0:
  testfailed("Wrong number of acquired vessels after releasing all vessels.")





"""
Custom pipeline functions used in Seattle Clearinghouse.

"""
from django.http import HttpResponseRedirect
from social_auth.backends.exceptions import AuthException
from social_auth.backends.pipeline.social import social_auth_user
from seattlegeni.website.control import interface
from uuid import uuid4
                                
def redirect_to_auto_register(*args, **kwargs):
    if not kwargs['request'].session.get('saved_username') and \
       kwargs.get('user') is None:
        return HttpResponseRedirect('/html/auto_register')

def username(request, *args, **kwargs):
    if kwargs.get('user'):
        username = kwargs['user'].username
    else:
        username = request.session.get('saved_username')
    return {'username': username}
   
def custom_social_auth_user(*args, **kwargs):
    try:
        return social_auth_user(*args, **kwargs)
    except AuthException:
    	 return HttpResponseRedirect('associate_error')      
    
def custom_create_user(backend, details, response, uid, username, user=None, *args,
                **kwargs):
    """Create user. Depends on get_username pipeline."""
    if user:
        return {'user': user}
    if not username:
        return None
    
    #set a random password
    password=str(uuid4())
    # NOTE: not return None because Django raises exception of strip email
    email = details.get('email') or ''
    if email == '':
      email= 'auto-register@'+ backend.name +'.com'
    user = interface.register_user(username, password, email, affiliation='auto-register@'+ backend.name)
    return {
        'user': user,
        'is_new': True
    }             



"""
<Program Name>
  vessels.py

<Started>
  July 16, 2009

<Author>
  Justin Samuel

<Purpose>
  Provides utilities for the controller interface (interface.py) to use for
  performing acquisition of different types of vessels as well as release of
  vessels.
"""

import traceback

from seattlegeni.common.exceptions import *

from seattlegeni.common.api import backend
from seattlegeni.common.api import lockserver
from seattlegeni.common.api import maindb

from seattlegeni.common.util import log
from seattlegeni.common.util import parallel

from seattlegeni.common.util.decorators import log_function_call





@log_function_call
def _parallel_process_vessels_from_list(vessel_list, process_func, lockserver_handle, *args):
  """
  Obtain locks on all of the nodes of vessels in vessel_list, get fresh vessel
  objects from the databae, and then parallelize a call to process_func to
  process each vessel in vessel_list (passing the additional *args to
  process_func).
  """
  
  node_id_list = []
  for vessel in vessel_list:
    node_id = maindb.get_node_identifier_from_vessel(vessel)
    # Lock names must be unique, and there could be multiple vessels from the
    # same node in the vessel_list.
    if node_id not in node_id_list:
      node_id_list.append(node_id)

  # Lock the nodes that these vessels are on.
  lockserver.lock_multiple_nodes(lockserver_handle, node_id_list)
  try:
    # Get new vessel objects from the db now that we have node locks.
    new_vessel_list = []
    for vessel in vessel_list:
      node_id = maindb.get_node_identifier_from_vessel(vessel)
      new_vessel_list.append(maindb.get_vessel(node_id, vessel.name))
    # Have the list object the caller may still be using contain the actual
    # vessel objects we have processed. That is, we've just replaced the
    # caller's list's contents with new vessel objects for the same vessels.
    vessel_list[:] = new_vessel_list[:]
  
    return parallel.run_parallelized(vessel_list, process_func, *args)
    
  finally:
    # Unlock the nodes.
    lockserver.unlock_multiple_nodes(lockserver_handle, node_id_list)





@log_function_call
def acquire_specific_vessels_best_effort(lockserver_handle, geniuser, vessel_list):
  """
  <Purpose>
    Acquire for geniuser as many vessels in vessel_list as possible.
  <Arguments>
    lockserver_handle
      The lockserver handle to be used for obtaining node locks.
    geniuser
      The GeniUser the vessels should be acquired for.
    vessel_list
      The vessels to attempt to acquire for geniuser.
  <Exceptions>
    None
  <Side Effects>
    Zero or more of the vessels are acquired for the user. The database has
    been updated to reflect the acquisition.
  <Returns>
    A list of the vessels that were acquired.
  """
  
  acquired_vessels = []
  
  parallel_results = _parallel_process_vessels_from_list(vessel_list, _do_acquire_vessel,
                                                         lockserver_handle, geniuser)

  # The "exception" key contains a list of tuples where the first item of
  # the tuple is the vessel object and the second item is the str(e) of
  # the exception. Because the repy parellelization module that is used
  # underneath only passes up the exception string, we have made
  # _do_acquire_vessel() include the string "UnableToAcquireResourcesError"
  # in the exception message so we can tell these apart from more
  # serious failures (e.g the backed is down).
  for (vessel, exception_message) in parallel_results["exception"]:
    
    if "UnableToAcquireResourcesError" in exception_message:
      # This is ok, maybe the node is offline.
      log.info("Failed to acquire vessel: " + str(vessel))
      
    elif "UnableToAcquireResourcesError" not in exception_message:
      # Something serious happened, maybe the backend is down.
      raise InternalError("Unexpected exception occurred during parallelized " + 
                          "acquisition of vessels: " + exception_message)
    
  # The "returned" key contains a list of tuples where the first item of
  # the tuple is the vessel object and the second is the return value
  # (which is None).
  for (ignored_argument_vessel, returned_vessel) in parallel_results["returned"]:
    # We successfully acquired this vessel.
    # Append the returned vessel from _do_acquire_vessel() rather than
    # the argument that the parallelize.r2py library used. Somewhere
    # along the way a copy of the argument_vessel is being made so it
    # doesn't reflect changes made to it.
    acquired_vessels.append(returned_vessel)

  return acquired_vessels





@log_function_call
def acquire_wan_vessels(lockserver_handle, geniuser, vesselcount):
  """
  <Purpose>
    Acquire 'wan' vessels for a geniuser.
  <Arguments>
    lockserver_handle
      The lockserver handle to be used for obtaining node locks.
    geniuser
      The GeniUser the vessels should be acquired for.
    vesselcount
      The number of vessels to acquire.
  <Exceptions>
    UnableToAcquireResourcesError
      If either the user does not not have enough vessel credits to acquire the
      number of vessels they requested or if there are not enough vessels
      available to fulfill the request.
  <Side Effects>
    The vessels are acquired for the user. The database has been updated to
    reflect the acquisition.
  <Returns>
    A list of the vessels that were acquired.
  """
  
  # Get a randomized list of vessels where no two vessels are on the same subnet.
  vessel_list = maindb.get_available_wan_vessels(geniuser, vesselcount)
  
  return _acquire_vessels_from_list(lockserver_handle, geniuser, vesselcount, vessel_list)





@log_function_call
def acquire_nat_vessels(lockserver_handle, geniuser, vesselcount):
  """
  <Purpose>
    Acquire 'nat' vessels for a geniuser.
  <Arguments>
    lockserver_handle
      The lockserver handle to be used for obtaining node locks.
    geniuser
      The GeniUser the vessels should be acquired for.
    vesselcount
      The number of vessels to acquire.
  <Exceptions>
    UnableToAcquireResourcesError
      If either the user does not not have enough vessel credits to acquire the
      number of vessels they requested or if there are not enough vessels
      available to fulfill the request.
  <Side Effects>
    The vessels are acquired for the user. The database has been updated to
    reflect the acquisition.
  <Returns>
    A list of the vessels that were acquired.
  """
  
  # Get a randomized list of nat vessels.
  vessel_list = maindb.get_available_nat_vessels(geniuser, vesselcount)
  
  return _acquire_vessels_from_list(lockserver_handle, geniuser, vesselcount, vessel_list)





@log_function_call
def acquire_lan_vessels(lockserver_handle, geniuser, vesselcount):
  """
  <Purpose>
    Acquire 'lan' vessels for a geniuser.
  <Arguments>
    lockserver_handle
      The lockserver handle to be used for obtaining node locks.
    geniuser
      The GeniUser the vessels should be acquired for.
    vesselcount
      The number of vessels to acquire.
  <Exceptions>
    UnableToAcquireResourcesError
      If either the user does not not have enough vessel credits to acquire the
      number of vessels they requested or if there are not enough vessels
      available to fulfill the request.
  <Side Effects>
    The vessels are acquired for the user. The database has been updated to
    reflect the acquisition.
  <Returns>
    A list of the vessels that were acquired.
  """
  
  # Get a randomized list that itself contains lists of vessels on the same subnet.
  subnet_vessel_list = maindb.get_available_lan_vessels_by_subnet(geniuser, vesselcount)
  
  # This case is a little more involved than with wan or rand vessels. If we
  # fail to get the number of desired vessels from one subnet, we need to try
  # another until we are out of subnets to try.
  for vessel_list in subnet_vessel_list:
    try:
      # If we don't hit an exception and return, then we found a subnet where
      # we could acquire all of the requested vessels. So, we're done.
      return _acquire_vessels_from_list(lockserver_handle, geniuser, vesselcount, vessel_list)
    except UnableToAcquireResourcesError:
      # Try the next subnet.
      continue

  # If we made it here, we tried all subnets in our list.
  raise UnableToAcquireResourcesError





@log_function_call
def acquire_rand_vessels(lockserver_handle, geniuser, vesselcount):
  """
  <Purpose>
    Acquire 'rand' vessels for a geniuser.
  <Arguments>
    lockserver_handle
      The lockserver handle to be used for obtaining node locks.
    geniuser
      The GeniUser the vessels should be acquired for.
    vesselcount
      The number of vessels to acquire.
  <Exceptions>
    UnableToAcquireResourcesError
      If either the user does not not have enough vessel credits to acquire the
      number of vessels they requested or if there are not enough vessels
      available to fulfill the request.
  <Side Effects>
    The vessels are acquired for the user. The database has been updated to
    reflect the acquisition.
  <Returns>
    A list of the vessels that were acquired.
  """
  
  # Get a randomized list of vessels where there are no guarantees about whether
  # the list includes wan vessels, lan vessels, vessels on the same subnet, etc.
  vessel_list = maindb.get_available_rand_vessels(geniuser, vesselcount)
  
  return _acquire_vessels_from_list(lockserver_handle, geniuser, vesselcount, vessel_list)





@log_function_call
def _acquire_vessels_from_list(lockserver_handle, geniuser, vesselcount, vessel_list):
  """
  This function will try to acquire vesselcount vessels from vessel_list.
  If less than vesselcount can be acquired, then the partial set of
  vessels that were acquired will be released by this function before it
  returns.
  
  Returns the list of acquired vessels if successful.
  """
  
  # Make sure there are sufficient vessels to even try to fulfill the request.
  if len(vessel_list) < vesselcount:
    raise UnableToAcquireResourcesError("There are not enough available vessels to fulfill the request.")
  
  acquired_vessels = []

  remaining_vessel_list = vessel_list[:]

  # Keep trying to acquire vessels until there are no more left to acquire.
  # There's a "return" statement in the loop that will get out of the loop
  # once we've obtained all of the vessels we wanted, so here we are only
  # concerned with there being any vessels left to try.
  while len(remaining_vessel_list) > 0:
  
    # Each time through the loop we'll try to acquire the number of vessels
    # remaining that are needed to fulfill the user's request.
    remaining_needed_vesselcount = vesselcount - len(acquired_vessels)
    next_vessels_to_acquire = remaining_vessel_list[:remaining_needed_vesselcount]
    remaining_vessel_list = remaining_vessel_list[remaining_needed_vesselcount:]
  
    # Note that we haven't worried about checking if the number of remaining
    # vessels could still fulfill the user's request. In the name of
    # correctness over efficiency, we'll let this case that should be rare
    # (at least until the vesselcount's users are request get to be huge)
    # sort itself out with a few unnecessary vessel acquisition before they
    # ultimately get released after this loop.
  
    parallel_results = _parallel_process_vessels_from_list(next_vessels_to_acquire, _do_acquire_vessel,
                                                           lockserver_handle, geniuser)
  
    # The "exception" key contains a list of tuples where the first item of
    # the tuple is the vessel object and the second item is the str(e) of
    # the exception. Because the repy parellelization module that is used
    # underneath only passes up the exception string, we have made
    # _do_acquire_vessel() include the string "UnableToAcquireResourcesError"
    # in the exception message so we can tell these apart from more
    # serious failures (e.g the backed is down).
    for (vessel, exception_message) in parallel_results["exception"]:
      
      if "UnableToAcquireResourcesError" in exception_message:
        # This is ok, maybe the node is offline.
        log.info("Failed to acquire vessel: " + str(vessel))
        
      elif "UnableToAcquireResourcesError" not in exception_message:
        # Something serious happened, maybe the backend is down.
        raise InternalError("Unexpected exception occurred during parallelized " + 
                            "acquisition of vessels: " + exception_message)
    
    # The "returned" key contains a list of tuples where the first item of
    # the tuple is the vessel object and the second is the return value
    # (which is None).
    for (ignored_argument_vessel, returned_vessel) in parallel_results["returned"]:
      # We successfully acquired this vessel.
      # Append the returned vessel from _do_acquire_vessel() rather than
      # the argument that the parallelize.r2py library used. Somewhere
      # along the way a copy of the argument_vessel is being made so it
      # doesn't reflect changes made to it.
      acquired_vessels.append(returned_vessel)

    # If we've acquired all of the vessels the user wanted, we're done.
    if len(acquired_vessels) == vesselcount:
      log.info("Successfully acquired vessel: " + str(returned_vessel))
      return acquired_vessels

  # If we got here, then we didn't acquire the vessels the user wanted. We
  # release any vessels that may have been acquired rather than leave the user
  # with a partial set of what they requested.
  if acquired_vessels:
    release_vessels(lockserver_handle, geniuser, acquired_vessels)

  raise UnableToAcquireResourcesError("Failed to acquire enough vessels to fulfill the request")





@log_function_call
def _do_acquire_vessel(vessel, geniuser):
  """
  Perform that actual acquisition of the vessel by the user (through the
  backend) and update the database accordingly if the vessel is successfully
  acquired.
  
  This gets called parallelized after a node lock is already obtained for the
  vessel.

  When an UnableToAcquireResourcesError is raised, the exception message
  will contain the string "UnableToAcquireResourcesError" so that it can be
  seen in the results of a call to repy's parallelization function.  
  """
  
  node_id = maindb.get_node_identifier_from_vessel(vessel)
  
  if vessel.acquired_by_user is not None:
    message = "Vessel already acquired once the node lock was obtained."
    raise UnableToAcquireResourcesError("UnableToAcquireResourcesError: " + message)
  
  node = maindb.get_node(node_id)
  if node.is_active is False:
    message = "Vessel's node is no longer active once the node lock was obtained."
    raise UnableToAcquireResourcesError("UnableToAcquireResourcesError: " + message)
  
  # This will raise a UnableToAcquireResourcesException if it fails (e.g if
  # the node is down). We want to allow the exception to be passed up to
  # the caller.
  try:
    backend.acquire_vessel(geniuser, vessel)
  except UnableToAcquireResourcesError, e:
    raise UnableToAcquireResourcesError("UnableToAcquireResourcesError: " + str(e))
  
  # Update the database to reflect the successful vessel acquisition.
  maindb.record_acquired_vessel(geniuser, vessel)
    
  # The _acquire_vessels_from_list() function will make use of this return value.
  return vessel





@log_function_call
def flag_vessels_for_user_keys_sync(lockserver_handle, vessel_list):
  """
  This function will mark (flag) the vessels in vessel_list as needing
  their user keys synced. The backend will then notice the vessels flagged
  in this way and will sync the user keys.
  
  We don't try to do the actual updating of user keys on vessels as that seems
  like a bad user experience, especially down the road when we may allow the
  users some form of giving others access to their vessels. Imagine a user wants
  to add ten users to their 300 acquired vessels and every time they add one
  of the users they have to wait minutes for keys to be update on vessels, and
  then we still get stuck having to deal with the case of temporary failures
  in communication and needing to go back and update some vessels later.
  
  <Returns>
    None
  """
  
  # This is going to be a potentially large list of vessels to obtain node
  # locks on simultaneously. This is something to revisit if lock contention
  # starts to become an issue.
  parallel_results = _parallel_process_vessels_from_list(vessel_list, _do_flag_vessels_for_user_keys_sync,
                                                         lockserver_handle)
  
  for (vessel, exception_message) in parallel_results["exception"]:
    raise InternalError("Unexpected exception occurred during parallelized " + 
                        "vessel user key out-of-sync flagging: " + exception_message)
  




@log_function_call
def _do_flag_vessels_for_user_keys_sync(vessel):
  """
  Indicates in the database that the vessel needs its user keys sync'd. This
  gets called parallelized after a node lock is already obtained for the
  vessel.
  """
  
  maindb.mark_vessel_as_needing_user_key_sync(vessel)





@log_function_call
def release_vessels(lockserver_handle, geniuser, vessel_list):
  """
  <Purpose>
    Release vessels (regardless of which user has acquired them)
  <Arguments>
    lockserver_handle
      The lockserver handle to use for acquiring node locks.
    geniuser
      The geniuser that has acquired the vessels to be released.
    vessel_list
      A list of vessels to be released.
  <Exceptions>
    None.
  <Side Effects>
    The vessels in the vessel_list are released.
  <Returns>
    None.
  """
    
  # This is going to be a potentially large list of vessels to obtain node
  # locks on simultaneously. This is something to revisit if lock contention
  # starts to become an issue.
  parallel_results = _parallel_process_vessels_from_list(vessel_list, _do_release_vessel,
                                                         lockserver_handle, geniuser)
  
  for (vessel, exception_message) in parallel_results["exception"]:
      raise InternalError("Unexpected exception occurred during parallelized " + 
                          "release of vessels: " + exception_message)

    



@log_function_call
def _do_release_vessel(vessel, geniuser):
  """
  Obtains a lock on the node the vessel is on and then makes a call to the
  backend to release the vessel.
  """
  
  if vessel.acquired_by_user != geniuser:
    # The vessel was either already released, someone is trying to do things
    # they shouldn't, or we have a bug.
    log.info("Not releasing vessel " + str(vessel) + 
             " because it is not acquired by user " + str(geniuser))
    return
  
  # We don't check for node.is_active == True because we might as well have
  # the backend try to clean up the vessel even if the database says it's
  # inactive (maybe the node is back online?).
  
  # This will not raise an exception, even if the node the vessel is on is down.
  backend.release_vessel(vessel)
  
  # Update the database to reflect the release of the vessel.
  maindb.record_released_vessel(vessel)





@log_function_call
def renew_vessels(lockserver_handle, geniuser, vessel_list):
  """
  <Purpose>
    Renew vessels.
  <Arguments>
    lockserver_handle
      The lockserver handle to use for acquiring node locks.
    geniuser
      The user who has acquired the vessels.
    vessel_list
      A list of vessels to be renewed.
  <Exceptions>
    InvalidRequestError
      If any vessels in vessel_list are not already acquired by geniuser.
  <Side Effects>
    The vessels in the vessel_list have expirations dates which are the maximum
    length of time from now that we allow renewal for.
  <Returns>
    None.
  """
  
  # This is going to be a potentially large list of vessels to obtain node
  # locks on simultaneously. This is something to revisit if lock contention
  # starts to become an issue.
  parallel_results = _parallel_process_vessels_from_list(vessel_list, _do_renew_vessel,
                                                         lockserver_handle, geniuser)
  
  for (vessel, exception_message) in parallel_results["exception"]:
      raise InternalError("Unexpected exception occurred during parallelized " + 
                          "renewal of vessels: " + exception_message)
  




@log_function_call
def _do_renew_vessel(vessel, geniuser):
  
  if vessel.acquired_by_user != geniuser:
    # The vessel was either already released, someone is trying to do things
    # they shouldn't, or we have a bug.
    log.info("Not renewing vessel " + str(vessel) + 
             " because it is not acquired by user " + str(geniuser))
    return
    
  maindb.set_maximum_vessel_expiration(vessel)





# The seattlegeni testlib must be imported first. We do this in case this
# module gets executed as a test to prevent it from "failing".
from seattlegeni.tests import testlib

import random

from seattlegeni.common.api import backend
from seattlegeni.common.api import lockserver
from seattlegeni.common.api import nodemanager
from seattlegeni.common.api import maindb

from seattlegeni.common.exceptions import *

from seattlegeni.node_state_transitions import node_transition_lib

from seattle.repyportability import *
add_dy_support(locals())

dy_import_module_symbols("rsa.r2py")




testusername = "testuser"

node_ip = "127.0.0.1"
node_port = 1224
node_address = node_ip + ":" + str(node_port)

nodeid_key = {"e" : 1, "n" : 2}
nodeid_key_str = str(nodeid_key["e"]) + " " + str(nodeid_key["n"])

donor_key = {"e" : 3, "n" : 4}
donor_key_str = str(donor_key["e"]) + " " + str(donor_key["n"])

per_node_key = {"e" : 5, "n" : 6}
per_node_key_str = str(per_node_key["e"]) + " " + str(per_node_key["n"])

extra_vessel_name = "v1"

def mockutil_cleanup_variables():
  global testusername
  global node_ip
  global node_port
  global node_address
  global nodeid_key
  global nodeid_key_str
  global donor_key
  global donor_key_str
  global per_node_key
  global per_node_key_str
  global extra_vessel_name

  testusername = "testuser"

  node_ip = "127.0.0.1"
  node_port = 1224
  node_address = node_ip + ":" + str(node_port)

  nodeid_key = {"e" : 1, "n" : 2}
  nodeid_key_str = str(nodeid_key["e"]) + " " + str(nodeid_key["n"])

  donor_key = {"e" : 3, "n" : 4}
  donor_key_str = str(donor_key["e"]) + " " + str(donor_key["n"])

  per_node_key = {"e" : 5, "n" : 6}
  per_node_key_str = str(per_node_key["e"]) + " " + str(per_node_key["n"])

  extra_vessel_name = "v1"

def mock_transitionlib_do_advertise_lookup(nodeaddress_list_to_return):
  
  def _mock_do_advertise_lookup(state):
    return nodeaddress_list_to_return
  
  node_transition_lib._do_advertise_lookup = _mock_do_advertise_lookup





def mock_nodemanager_get_node_info(nodeid_key, version, vessels_dict):
  
  def _mock_get_node_info(ip, port):
    nodeinfo = {"version" : version,
                "nodename" : "",
                "nodekey" : nodeid_key,
                "vessels" : {}}
    nodeinfo["vessels"] = vessels_dict
    return nodeinfo
  
  nodemanager.get_node_info = _mock_get_node_info





def mock_lockserver_calls():
  
  def _mock_create_lockserver_handle(lockserver_url=None):
    pass
  lockserver.create_lockserver_handle = _mock_create_lockserver_handle
  
  def _mock_destroy_lockserver_handle(lockserver_handle):
    pass
  lockserver.destroy_lockserver_handle = _mock_destroy_lockserver_handle
  
  def _mock_perform_lock_request(request_type, lockserver_handle, user_list=None, node_list=None):
    pass
  lockserver._perform_lock_request = _mock_perform_lock_request




def mock_backend_generate_key(keylist):

  def _mock_generate_key(keydescription):
    return keylist.pop()
  
  backend.generate_key = _mock_generate_key



def mock_nodemanager_get_vesselresources():
  
  def _mock_get_vessel_resources(*args):
    resource_dict={}
    resourcelist = []
    #generate resource list
    for i in range(90):
      resourcelist.append(i)
    resource_dict['usableports'] = resourcelist
    
    return resource_dict

  nodemanager.get_vessel_resources = _mock_get_vessel_resources




# How many times backend.set_vessel_owner_key() has been called.
set_vessel_owner_key_call_count = 0

def mock_backend_set_vessel_owner_key():

  global set_vessel_owner_key_call_count

  def _mock_set_vessel_owner_key(node, vesselname, old_ownerkey, new_ownerkey):

    global set_vessel_owner_key_call_count

    print "[_mock_set_vessel_owner_key] called: ", node, vesselname, old_ownerkey, new_ownerkey

    # We changed things so that the node will now be inactive until the node
    # is changed to the onepercentmanyevents state.
    #assert(node == maindb.get_active_nodes()[0])
    assert(vesselname == extra_vessel_name)
    assert(old_ownerkey == donor_key_str)
    assert(new_ownerkey == per_node_key_str)

    set_vessel_owner_key_call_count += 1
  
  set_vessel_owner_key_call_count = 0

  backend.set_vessel_owner_key = _mock_set_vessel_owner_key




# How many times backend.set_vessel_user_keylist() has been called.
set_vessel_user_keylist_call_count = 0

def mock_backend_set_vessel_user_keylist(expected_key_list):

  global set_vessel_user_keylist_call_count

  def _mock_set_vessel_user_keylist(node, vesselname, userkeylist):

    global set_vessel_user_keylist_call_count

    print "[_mock_set_vessel_user_keylist] called: ", node, vesselname, userkeylist

    # We changed things so that the node will now be inactive until the node
    # is changed to the onepercentmanyevents state.
    #assert(node == maindb.get_active_nodes()[0])

    # Make sure that the name of the vessel is correct, if we are setting
    # the keylist for the extra vessel.

    if userkeylist != []:
      assert(vesselname == extra_vessel_name)

    assert(userkeylist == expected_key_list or userkeylist == [])

    set_vessel_user_keylist_call_count += 1

  set_vessel_user_keylist_call_count = 0

  backend.set_vessel_user_keylist = _mock_set_vessel_user_keylist




#How many times backend.split_vessel() has been called
split_vessel_call_count = 0

def mock_backend_split_vessel():

  global split_vessel_call_count

  def _mock_split_vessel(node, vesselname, resource_data):

    global split_vessel_call_count
    global extra_vessel_name

    print "[_mock_split_vessel] called: ", node, vesselname, resource_data

    # We changed things so that the node will now be inactive until the node
    # is changed to the onepercentmanyevents state.
    #assert(node == maindb.get_active_nodes()[0])
    assert(vesselname == extra_vessel_name)

    split_vessel_call_count += 1
    extra_vessel_name = "extra_vessel_split"+str(split_vessel_call_count)
    return ("extra_vessel_split"+str(split_vessel_call_count), 
            "new_vessel"+str(split_vessel_call_count))

  split_vessel_call_count = 0

  backend.split_vessel = _mock_split_vessel




#How many times backend.join_vessels() has been called
join_vessels_call_count = 0

def mock_backend_join_vessels():

  global join_vessels_call_count

  def _mock_join_vessels(node, extra_vesselname, other_vesselname):

    global join_vessels_call_count
    global extra_vessel_name

    print "[_mock_join_vessels] called: ", node, extra_vesselname, other_vesselname

    # We changed things so that the node will now be inactive until the node
    # is changed to the onepercentmanyevents state.
    #assert(node == maindb.get_active_nodes()[0])
    assert(extra_vesselname == extra_vessel_name)

    join_vessels_call_count += 1

    extra_vessel_name = "extra_vessel_join"+str(join_vessels_call_count)

    return extra_vessel_name

  join_vessels_call_count = 0

  backend.join_vessels = _mock_join_vessels




#mock up for pubkey to string
def _mock_pubkey_to_string(pubkey):
  pub_string = str(pubkey['e'])+" "+str(pubkey['n'])
  return pub_string

node_transition_lib._do_rsa_publickey_to_string = _mock_pubkey_to_string



#!/usr/bin/env python

from seattlegeni.backend import backend_daemon
from seattlegeni.common.api import keydb
from seattlegeni.common.api import nodemanager

from seattlegeni.common.util.decorators import log_function_call

@log_function_call
def _mock_do_signed_call(nodehandle, *callargs):
  pass
nodemanager._do_signed_call = _mock_do_signed_call

@log_function_call
def mock_get_private_key(pubkey):
  return 'Mock private key'
keydb.get_private_key = mock_get_private_key

@log_function_call
def mock_set_private_key(pubkey, privkey, keydescription):
  pass
keydb.set_private_key = mock_set_private_key

# Start the backend.
backend_daemon.main()




"""
<Program>
  validations.py

<Started>
  29 June 2009

<Author>
  Justin Samuel

<Purpose>
  This is a utility module to provide functions to valid user input for specific
  situations such as account registration. These methods raise a
  ValidationError if the value being checked is invalid. The message of the
  exceptions --- which is accessible with, for example str(e) --- will contain
  publicly-displayable information about why the validation failed. If the
  validation succeeds, nothing is returned.
"""

import re

from seattlegeni.common.exceptions import *

from seattlegeni.common.util.assertions import *

from seattle.repyportability import *
add_dy_support(locals())

dy_import_module_symbols("rsa.r2py")






USERNAME_MIN_LENGTH = 3
USERNAME_MAX_LENGTH = 40
# In general, numbers, letters, and underscores are allowed in usernames.
USERNAME_ALLOWED_CHARS = r"0-9a-zA-Z_"
USERNAME_ALLOWED_REGEX = re.compile(r"^[" + USERNAME_ALLOWED_CHARS + "]+$")
# The first character of the username, however, we don't allow to be an underscore.
# We make this a separate regex so that it's easier to explain to the user
# if their username is denied.
USERNAME_ALLOWED_FIRST_CHARS = r"0-9a-zA-Z"
USERNAME_ALLOWED_FIRST_REGEX = re.compile(r"^[" + USERNAME_ALLOWED_FIRST_CHARS + "]")

PASSWORD_MIN_LENGTH = 6

# This regex is borrowed from http://www.regular-expressions.info/email.html
# It could probably be more restrictive and we'd be fine, but, really, if we're
# looking to ensure valid email addresses then we need to send confirmation
# emails.
EMAIL_VALID_REGEX = re.compile(r"[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*@(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?")

AFFILIATION_MIN_LENGTH = 3
AFFILIATION_MAX_LENGTH = 200





def validate_username(username):
  """
  <Purpose>
    Determine whether username is a valid username.
  """
  try:
    assert_str(username)
  except AssertionError:
    raise ValidationError("Username must be a string.")
  
  if len(username) < USERNAME_MIN_LENGTH:
    raise ValidationError("Username must be at least " + str(USERNAME_MIN_LENGTH) + " characters.")
  
  if len(username) > USERNAME_MAX_LENGTH:
    raise ValidationError("Username must be less than " + str(USERNAME_MAX_LENGTH) + " characters.")
  
  if not USERNAME_ALLOWED_REGEX.match(username):
    raise ValidationError("Username can only contain the characters " + USERNAME_ALLOWED_CHARS)

  if not USERNAME_ALLOWED_FIRST_REGEX.match(username):
    raise ValidationError("Username must start with one of the characters " + USERNAME_ALLOWED_FIRST_CHARS)





def validate_password(password):
  """
  <Purpose>
    Determine whether password is a valid (strong enough) password. The code that
    calls this function should probably also make sure that the password isn't
    the same as the username.
  """
  try:
    assert_str(password)
  except AssertionError:
    raise ValidationError("Password must be a string.")
  
  if len(password) < PASSWORD_MIN_LENGTH:
    raise ValidationError("Password must be at least " + str(PASSWORD_MIN_LENGTH) + " characters.")





def validate_username_and_password_different(username, password):
  """
  <Purpose>
    Determine whether username is different enough from the password.
  """
  try:
    assert_str(username)
    assert_str(password)
  except AssertionError:
    raise ValidationError("Username and password must be strings.")
  
  if username == password:
    raise ValidationError("Username cannot be the same as the password.")





def validate_email(email):
  """
  <Purpose>
    Determine whether email is a valid email address.
  """
  try:
    assert_str(email)
  except AssertionError:
    raise ValidationError("E-mail must be a string.")
  
  if not EMAIL_VALID_REGEX.match(email):
    raise ValidationError("E-mail address is not valid.")





def validate_affiliation(affiliation):
  """
  <Purpose>
    Determine whether affiliation is a valid affiliation.
  """
  try:
    assert_str(affiliation)
  except AssertionError:
    raise ValidationError("Affiliation must be a string.")
  
  if len(affiliation) < AFFILIATION_MIN_LENGTH:
    raise ValidationError("Affiliation must be at least " + str(AFFILIATION_MIN_LENGTH) + " characters.")
  
  if len(affiliation) > AFFILIATION_MAX_LENGTH:
    raise ValidationError("Affiliation must be less than " + str(AFFILIATION_MAX_LENGTH) + " characters.")





def validate_pubkey_string(pubkeystring):
  """
  <Purpose>
    Determine whether a pubkey string looks like a valid public key.
  <Arguments>
    pubkeystring
      The string we want to find out if it is a valid pubkey string.
  <Exceptions>
    None
  <Side Effects>
    None
  <Returns>
    True if pubkeystring looks like a valid pubkey, False otherwise.
    Currently this uses functions that aren't conclusive about validity of the
    key, but for our purposes we just want to make sure it's generally the
    correct data format.
  """
  try:
    assert_str(pubkeystring)
  except AssertionError:
    raise ValidationError("Public key must be a string.")
  
  try:
    possiblepubkey = rsa_string_to_publickey(pubkeystring)
  except ValueError:
    raise ValidationError("Public key is not of a correct format.")

  if not rsa_is_valid_publickey(possiblepubkey):
    raise ValidationError("Public key is invalid.")





#!/usr/bin/env python
"""
Print a summarization of the status of seattlegeni, according to information in
the database.
"""

from seattlegeni.common.util import statistics






if __name__ == "__main__":
  
  # Decrease the amount of logging output.
  from seattlegeni.common.util import log
  log.loglevel = log.LOG_LEVEL_CRITICAL
  
  
  
  print "Available vessels by port"
  print "-------------------------"
  
  available_vessels_dict = statistics.get_available_vessel_counts_by_port()
  
  print "%s\t%s\t%s\t%s" % ("Port", "All", "No-NAT", "Only-NAT")
  print "%s\t%s\t%s\t%s" % ("----", "---", "------", "--------")
  
  for port in available_vessels_dict:
    portdict = available_vessels_dict[port]
    print "%s\t%s\t%s\t%s" % (port, portdict["all"], portdict["no_nat"], portdict["only_nat"])



  print
  print "Number of vessels acquired per user"
  print "-----------------------------------"
  
  vessel_acquisition_dict = statistics.get_vessel_acquisition_counts_by_user()
  
  for username in vessel_acquisition_dict:
    print username + (" " * (25 - len(username))) + str(vessel_acquisition_dict[username])
  
  print
  print "Vessel acquisition summary"
  print "--------------------------"
  print "There are " +  str(len(vessel_acquisition_dict.keys())) + " users who have acquired vessels."
  print "There are " + str(sum(vessel_acquisition_dict.values())) + " acquired vessels."



  print
  print "Donations per user"
  print "------------------"
  
  donations_dict = statistics.get_donation_counts_by_user()
  
  print "User" + (" " * (25 - len("User"))) + "Active" + "\t" + "Inactive"
  print "----" + (" " * (25 - len("----"))) + "------" + "\t" + "--------"
    
  total_active_donations = 0
  total_inactive_donations = 0
    
  for username in donations_dict:
    (active_donation_count, inactive_donation_count) = donations_dict[username]
    total_active_donations += active_donation_count 
    total_inactive_donations +=  inactive_donation_count
    print username + (" " * (25 - len(username))) + str(active_donation_count) + "\t" + str(inactive_donation_count)

  print
  print "Donation summary"
  print "--------------------------"
  print "Total active donations:   " +  str(total_active_donations)
  print "Total inactive donations: " +  str(total_inactive_donations)




"""
<Program>
  backend.py

<Started>
  29 June 2009

<Author>
  Justin Samuel

<Purpose>
  This is the API that should be used to interact with the backend XML-RPC
  server. Functions in this module are the only way that other code should
  interact with the backend.
   
  All components of seattlegeni use the backend to perform any nodemanager
  communication that changes the state of a node. ("State" here means
  modifying the node in any way, not "state" as in the "canonical state".)
  If the node only needs to be queried, not modified, then that can be
  done directly using the nodemanager api. This prevents the backend from
  needing to be called for regular querying done by polling daemons yet
  keeps the backend as the sole place where node changing operations
  are performed.
   
  The website uses only four functions in this module:
    acquire_vessel()
    release_vessel()
    generate_key()
    set_vessel_users()
     
  The other functions are used by polling daemons. In order to use the other
  functions, set_backend_authcode() must be called by the script first.
  
  Note that the none of these function calls will result in changes to the
  database. Any corresponding changes that also need to be made in the
  database must be made by any client code that uses this api.
"""

import socket
import traceback
import xmlrpclib

from seattlegeni.common.exceptions import *

from seattlegeni.common.util.decorators import log_function_call





BACKEND_URL = "http://127.0.0.1:8020"

# This is not a thread-local variable. There isn't any case where one thread
# in a process will have this authcode and another won't. So, we don't make
# the client code pass around a handle just for this.
backend_authcode = None





def _get_backend_proxy():
  return xmlrpclib.ServerProxy(BACKEND_URL)





def _do_backend_request(func, *args):
  try:
    return func(*args)
  except xmlrpclib.Fault, fault:
    if fault.faultCode == 100:
      raise NodemanagerCommunicationError(fault.faultString)
    else:
      raise ProgrammerError("The backend rejected the request: " + traceback.format_exc())
  except xmlrpclib.ProtocolError:
    raise InternalError("Unable to communicate with the backend: " + traceback.format_exc())
  except socket.error:
    raise InternalError("Unable to communicate with the backend: " + traceback.format_exc())






def _require_backend_authcode():
  if backend_authcode is None:
    raise ProgrammerError("You must call set_backend_authcode(authcode) before calling this function.")





# Not using log_function_call so we don't log the authcode.
def set_backend_authcode(authcode):
  """
  <Purpose>
    Sets the value of the authcode sent to the backend with privileged requests.
    This is needed for the backend to ensure that calls to the privileged
    operations (set_vessel_owner, split_vessel, join_vessel) are allowed.
    The website will never need to use this function (and shouldn't have
    access to a valid authcode, either). This function will need to be used
    by polling daemons such as node state transition scripts.
  <Arguments>
    authcode
      The authcode to send with privileged requests.
  <Exceptions>
    None
  <Side Effects>
    The value of the global variable backend_authcode has been changed.
  <Returns>
    None
  """
  global backend_authcode
  backend_authcode = authcode





@log_function_call
def acquire_vessel(geniuser, vessel):
  """
  <Purpose>
    Perform the necessary nodemanager communication to acquire a vessel for a
    user.
  <Arguments>
    geniuser
      The user the vessel is to be acquired for.
    vessel
      A Vessel object of the vessel to be acquired.
  <Exceptions>
    None
  <Side Effects>
    The vessel has been acquired for the user.
  <Returns>
    None
  """
  # Acquiring a vessel is just setting the userkeylist to include only this
  # user's key.
  func = _get_backend_proxy().SetVesselUsers
  args = (vessel.node.node_identifier, vessel.name, [geniuser.user_pubkey])
  
  try:
    _do_backend_request(func, *args)
  except NodemanagerCommunicationError:
    raise UnableToAcquireResourcesError





@log_function_call
def release_vessel(vessel):
  """
  <Purpose>
    Perform the necessary nodemanager communication to release a vessel.
  <Arguments>
    vessel
      A Vessel object of an acquired vessel to be released.
  <Exceptions>
    None
  <Side Effects>
    The vessel has been released (at least as far as the client code is
    concerned).
  <Returns>
    None
  """
  # This is actually a noop. The reason for this is because the backend doesn't
  # perform immediate action because of a released vessel. Instead, the client
  # code will user maindb.record_released_vessel() and the database change
  # resulting from that will cause the cleanup thread in the backend to do
  # anything it needs to do.
  pass





@log_function_call
def generate_key(keydecription):
  """
  <Purpose>
    Generate a new key pair through the backend. The backend will store the
    private key in the keydb. As of the time of writing this, this method
    is mainly to be used by the website to obtain a donor_pubkey without
    ever having to see the corresponding private key.
  <Arguments>
    keydecription
      A description of what this key is for. This should be specific enough
      to locate where the returned public key is stored in the maindb.
  <Exceptions>
    None
  <Side Effects>
    A new key has been generated and stored in the the keydb with the privided
    description.
  <Returns>
    The public key part of the key pair. The key is in string format.
  """
  func = _get_backend_proxy().GenerateKey
  args = (keydecription,)
  
  return _do_backend_request(func, *args)





@log_function_call
def set_vessel_user_keylist(node, vesselname, userkeylist):
  """
  <Purpose>
    Perform the necessary nodemanager communication to set the user key list
    for a vessel.
  <Arguments>
    node
      The Node object of the node which the vessel is on.
    vesselname
      The name of the vessel whose user key list is to be set.
    userkeylist
      A list of public key strings that are the user keys to be set for the vessel.
  <Exceptions>
    NodemanagerCommunicationError
      If there's a problem communicating with the node.
  <Side Effects>
    The user key list for the vessel has been changed on the vessel.
  <Returns>
    None
  """
  func = _get_backend_proxy().SetVesselUsers
  args = (node.node_identifier, vesselname, userkeylist)
  
  _do_backend_request(func, *args)





@log_function_call
def set_vessel_owner_key(node, vesselname, old_ownerkey, new_ownerkey):
  """
  <Purpose>
    Perform the necessary nodemanager communication to set the owner key
    for a vessel.
  <Arguments>
    node
      The Node object of the node which the vessel is on.
    vesselname
      The name of the vessel whose owner key is to be set.
    old_ownerkey
      A public key string of the owner key that is the current owner key for
      the vessel. Note that this key (with its correspond private key) must
      already exist in the keydb before this function is called.
    new_ownerkey
      A public key string of the owner key to be set for the vessel. Note that
      this key (with its correspond private key) must already exist in the
      keydb before this function is called.
  <Exceptions>
    NodemanagerCommunicationError
      If there's a problem communicating with the node.
  <Side Effects>
    The owner key on the vessel has been changed. The main database is not modified.
  <Returns>
    None
  """
  # Changing the owner key is a privileged request and the backend server will
  # require an authcode to be sent with the request.
  _require_backend_authcode()

  func = _get_backend_proxy().SetVesselOwner
  args = (backend_authcode, node.node_identifier, vesselname, old_ownerkey, new_ownerkey)
  
  _do_backend_request(func, *args)

  



@log_function_call
def split_vessel(node, vesselname, desiredresourcedata):
  """
  <Purpose>
    Perform the necessary nodemanager communication to split the vessel.
  <Arguments>
    node
      The Node object of the node which the vessel is on.
    vesselname
      The name of the vessel that is to be split.
    desiredresourcedata
      A string of resourcedata that specifies the resources of a new vessel to
      create when splitting the existing vessel. This resourcedata has the
      format of a resources file.
  <Exceptions>
    InvalidRequestError
      If unable to split the vessel. This includes if it fails because there
      aren't enough resources available to do the split.
    NodemanagerCommunicationError
      If there's a problem communicating with the node.
  <Side Effects>
    The vessel passed in as an argument to the function no longer exists.
    It has been split into two new vessels, one of which has the resources
    specified in the desiredresourcedata. The main database is not modified.
  <Returns>
    A tuple of the two new vessel names that resulted from the split. The
    first element of the tuple is the name of the vessel that has the
    leftover resources from the split. The second element of the tuple is
    the name of the vessel that has the exact resources specified in the
    desiredresourcedata.
  """
  # Splitting a vessel is a privileged request and the backend server will
  # require an authcode to be sent with the request.
  _require_backend_authcode()
  
  func = _get_backend_proxy().SplitVessel
  args = (backend_authcode, node.node_identifier, vesselname, desiredresourcedata)
  
  # We don't have a great way of having the backend tell us that the split
  # failed only due to lacking resources and not because of some other type
  # of problem. So, we're going to have to assume that all failures in this
  # call are because of not enough resources to do the split.
  try:
    return _do_backend_request(func, *args)
  except ProgrammerError, e:
    # The exception message will already contain the traceback info because
    # of the message that was set when the ProgrammerError was thrown in
    # the call to _do_backend_request().
    raise InvalidRequestError(str(e))





@log_function_call
def join_vessels(node, firstvesselname, secondvesselname):
  """
  <Purpose>
    Perform the necessary nodemanager communication to join the vessels.
  <Arguments>
    node
      The Node object of the node which the vessels are on.
    firstvesselname
      The name of the first vessel that is to be joined.
    secondvesselname
      The name of the second vessel that is to be joined.
  <Exceptions>
    NodemanagerCommunicationError
      If there's a problem communicating with the node.
  <Side Effects>
    The two vessel passed in as arguments to the function no longer exist.
    They have been joined into a new vessel. The first vessel will retain the
    user keys and therefore the state. The main database is not modified.
  <Returns>
    The name of the new vessel that has been created by joining together the
    two vessels.
  """
  # Joining two vessels is a privileged request and the backend server will
  # require an authcode to be sent with the request.
  _require_backend_authcode()
  
  func = _get_backend_proxy().JoinVessels
  args = (backend_authcode, node.node_identifier, firstvesselname, secondvesselname)
  
  return _do_backend_request(func, *args)
  



"""
This is the configuration file used by the keydb api. Any program that uses
the keydb must have access to this file. Any program that shouldn't have
access to the keydb should not have access to this file.
"""

# Fill these in with the information needed to use the key database.
dbuser = ""
dbpass = ""
dbname = "keydb"
dbhost = "localhost"
